import logging
import re
from typing import TYPE_CHECKING, List, cast
from unittest import mock

import pytest

from great_expectations.compatibility import azure
from great_expectations.core import IDDict
from great_expectations.core.batch import LegacyBatchDefinition
from great_expectations.core.partitioners import FileNamePartitionerPath
from great_expectations.core.util import AzureUrl
from great_expectations.datasource.fluent import BatchRequest
from great_expectations.datasource.fluent.data_connector import (
    AzureBlobStorageDataConnector,
)
from great_expectations.datasource.fluent.data_connector.file_path_data_connector import (
    MissingFilePathTemplateMapFnError,
)

if TYPE_CHECKING:
    from great_expectations.datasource.fluent.data_connector import (
        DataConnector,
    )


logger = logging.getLogger(__name__)


if not (azure.storage and azure.BlobServiceClient and azure.ContainerClient):  # type: ignore[truthy-function] # False if NotImported
    pytest.skip(
        'Could not import "azure.storage.blob" from Microsoft Azure cloud',
        allow_module_level=True,
    )


class MockContainerClient:
    pass


class MockBlobServiceClient:
    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def get_container_client(self, container: str) -> azure.ContainerClient:
        return cast("azure.ContainerClient", MockContainerClient())


@pytest.mark.big
@mock.patch(
    "great_expectations.datasource.fluent.data_asset.data_connector.azure_blob_storage_data_connector.list_azure_keys"
)
def test_basic_instantiation(mock_list_keys):
    mock_list_keys.return_value = [
        "alpha-1.csv",
        "alpha-2.csv",
        "alpha-3.csv",
    ]

    azure_client: azure.BlobServiceClient = cast("azure.BlobServiceClient", MockBlobServiceClient())
    my_data_connector: DataConnector = AzureBlobStorageDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_azure_blob_storage_data_asset",
        batching_regex=re.compile(r"alpha-(.*)\.csv"),
        azure_client=azure_client,
        account_name="my_account",
        container="my_container",
        name_starts_with="",
        file_path_template_map_fn=AzureUrl.AZURE_BLOB_STORAGE_HTTPS_URL_TEMPLATE.format,
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "alpha-1.csv",
        "alpha-2.csv",
        "alpha-3.csv",
    ]
    assert my_data_connector.get_matched_data_reference_count() == 3
    assert my_data_connector.get_matched_data_references()[:3] == [
        "alpha-1.csv",
        "alpha-2.csv",
        "alpha-3.csv",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert my_data_connector.get_unmatched_data_reference_count() == 0


@pytest.mark.big
@mock.patch(
    "great_expectations.datasource.fluent.data_asset.data_connector.azure_blob_storage_data_connector.list_azure_keys"
)
def test_instantiation_batching_regex_does_not_match_paths(mock_list_keys):
    mock_list_keys.return_value = [
        "alpha-1.csv",
        "alpha-2.csv",
        "alpha-3.csv",
    ]

    azure_client: azure.BlobServiceClient = cast("azure.BlobServiceClient", MockBlobServiceClient())
    my_data_connector: DataConnector = AzureBlobStorageDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_azure_blob_storage_data_asset",
        batching_regex=re.compile(r"(?P<name>.+)_(?P<timestamp>.+)_(?P<price>.*)\.csv"),
        azure_client=azure_client,
        account_name="my_account",
        container="my_container",
        name_starts_with="",
        file_path_template_map_fn=AzureUrl.AZURE_BLOB_STORAGE_HTTPS_URL_TEMPLATE.format,
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "alpha-1.csv",
        "alpha-2.csv",
        "alpha-3.csv",
    ]
    assert my_data_connector.get_matched_data_reference_count() == 0
    assert my_data_connector.get_matched_data_references()[:3] == []
    assert my_data_connector.get_unmatched_data_references()[:3] == [
        "alpha-1.csv",
        "alpha-2.csv",
        "alpha-3.csv",
    ]
    assert my_data_connector.get_unmatched_data_reference_count() == 3


@pytest.mark.big
@mock.patch(
    "great_expectations.datasource.fluent.data_asset.data_connector.azure_blob_storage_data_connector.list_azure_keys"
)
def test_return_all_batch_definitions_unsorted(mock_list_keys):
    mock_list_keys.return_value = [
        "abe_20200809_1040.csv",
        "alex_20200809_1000.csv",
        "alex_20200819_1300.csv",
        "eugene_20200809_1500.csv",
        "eugene_20201129_1900.csv",
        "james_20200713_1567.csv",
        "james_20200810_1003.csv",
        "james_20200811_1009.csv",
        "will_20200809_1002.csv",
        "will_20200810_1001.csv",
    ]

    azure_client: azure.BlobServiceClient = cast("azure.BlobServiceClient", MockBlobServiceClient())
    my_data_connector: DataConnector = AzureBlobStorageDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_azure_blob_storage_data_asset",
        batching_regex=re.compile(r"(?P<name>.+)_(?P<timestamp>.+)_(?P<price>.*)\.csv"),
        azure_client=azure_client,
        account_name="my_account",
        container="my_container",
        name_starts_with="",
        file_path_template_map_fn=AzureUrl.AZURE_BLOB_STORAGE_HTTPS_URL_TEMPLATE.format,
    )
    # with missing BatchRequest arguments
    with pytest.raises(TypeError):
        # noinspection PyArgumentList
        my_data_connector.get_batch_definition_list()

    # with empty options
    unsorted_batch_definition_list: List[LegacyBatchDefinition] = (
        my_data_connector.get_batch_definition_list(
            BatchRequest(
                datasource_name="my_file_path_datasource",
                data_asset_name="my_azure_blob_storage_data_asset",
                options={},
            )
        )
    )
    expected: List[LegacyBatchDefinition] = [
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_azure_blob_storage_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "abe_20200809_1040.csv",
                    "name": "abe",
                    "timestamp": "20200809",
                    "price": "1040",
                }
            ),
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_azure_blob_storage_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "alex_20200809_1000.csv",
                    "name": "alex",
                    "timestamp": "20200809",
                    "price": "1000",
                }
            ),
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_azure_blob_storage_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "alex_20200819_1300.csv",
                    "name": "alex",
                    "timestamp": "20200819",
                    "price": "1300",
                }
            ),
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_azure_blob_storage_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "eugene_20200809_1500.csv",
                    "name": "eugene",
                    "timestamp": "20200809",
                    "price": "1500",
                }
            ),
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_azure_blob_storage_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "eugene_20201129_1900.csv",
                    "name": "eugene",
                    "timestamp": "20201129",
                    "price": "1900",
                }
            ),
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_azure_blob_storage_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "james_20200713_1567.csv",
                    "name": "james",
                    "timestamp": "20200713",
                    "price": "1567",
                }
            ),
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_azure_blob_storage_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "james_20200810_1003.csv",
                    "name": "james",
                    "timestamp": "20200810",
                    "price": "1003",
                }
            ),
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_azure_blob_storage_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "james_20200811_1009.csv",
                    "name": "james",
                    "timestamp": "20200811",
                    "price": "1009",
                }
            ),
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_azure_blob_storage_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "will_20200809_1002.csv",
                    "name": "will",
                    "timestamp": "20200809",
                    "price": "1002",
                }
            ),
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_azure_blob_storage_data_asset",
            batch_identifiers=IDDict(
                {
                    "path": "will_20200810_1001.csv",
                    "name": "will",
                    "timestamp": "20200810",
                    "price": "1001",
                }
            ),
        ),
    ]
    assert expected == unsorted_batch_definition_list

    # with specified Batch query options
    unsorted_batch_definition_list = my_data_connector.get_batch_definition_list(
        BatchRequest(
            datasource_name="my_file_path_datasource",
            data_asset_name="my_azure_blob_storage_data_asset",
            options={"name": "alex", "timestamp": "20200819", "price": "1300"},
        )
    )
    assert expected[2:3] == unsorted_batch_definition_list


@pytest.mark.big
@mock.patch(
    "great_expectations.datasource.fluent.data_asset.data_connector.azure_blob_storage_data_connector.list_azure_keys"
)
def test_return_only_unique_batch_definitions(mock_list_keys):
    mock_list_keys.return_value = [
        "A/file_1.csv",
        "A/file_2.csv",
        "A/file_3.csv",
    ]

    azure_client: azure.BlobServiceClient = cast("azure.BlobServiceClient", MockBlobServiceClient())

    my_data_connector: DataConnector

    my_data_connector = AzureBlobStorageDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_azure_blob_storage_data_asset",
        batching_regex=re.compile(r"(?P<name>.+).*\.csv"),
        azure_client=azure_client,
        account_name="my_account",
        container="my_container",
        name_starts_with="A/",
        file_path_template_map_fn=AzureUrl.AZURE_BLOB_STORAGE_HTTPS_URL_TEMPLATE.format,
    )
    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "A/file_1.csv",
        "A/file_2.csv",
        "A/file_3.csv",
    ]
    assert my_data_connector.get_matched_data_reference_count() == 3
    assert my_data_connector.get_matched_data_references()[:3] == [
        "A/file_1.csv",
        "A/file_2.csv",
        "A/file_3.csv",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert my_data_connector.get_unmatched_data_reference_count() == 0

    mock_list_keys.return_value = [
        "B/file_1.csv",
        "B/file_2.csv",
    ]

    expected: List[LegacyBatchDefinition] = [
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_azure_blob_storage_data_asset",
            batch_identifiers=IDDict({"path": "B/file_1.csv", "filename": "file_1"}),
        ),
        LegacyBatchDefinition(
            datasource_name="my_file_path_datasource",
            data_connector_name="fluent",
            data_asset_name="my_azure_blob_storage_data_asset",
            batch_identifiers=IDDict({"path": "B/file_2.csv", "filename": "file_2"}),
        ),
    ]

    my_data_connector = AzureBlobStorageDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_azure_blob_storage_data_asset",
        batching_regex=re.compile(r"(?P<filename>.+).*\.csv"),
        azure_client=azure_client,
        account_name="my_account",
        container="my_container",
        name_starts_with="B/",
        file_path_template_map_fn=AzureUrl.AZURE_BLOB_STORAGE_HTTPS_URL_TEMPLATE.format,
    )

    unsorted_batch_definition_list: List[LegacyBatchDefinition] = (
        my_data_connector.get_batch_definition_list(
            BatchRequest(
                datasource_name="my_file_path_datasource",
                data_asset_name="my_azure_blob_storage_data_asset",
                options={},
            )
        )
    )
    assert expected == unsorted_batch_definition_list


@pytest.mark.big
@mock.patch(
    "great_expectations.datasource.fluent.data_asset.data_connector.azure_blob_storage_data_connector.list_azure_keys"
)
def test_alpha(mock_list_keys):
    mock_list_keys.return_value = [
        "test_dir_alpha/A.csv",
        "test_dir_alpha/B.csv",
        "test_dir_alpha/C.csv",
        "test_dir_alpha/D.csv",
    ]

    azure_client: azure.BlobServiceClient = cast("azure.BlobServiceClient", MockBlobServiceClient())
    my_data_connector: DataConnector = AzureBlobStorageDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_azure_blob_storage_data_asset",
        batching_regex=re.compile(r"(?P<part_1>.*)\.csv"),
        azure_client=azure_client,
        account_name="my_account",
        container="my_container",
        name_starts_with="test_dir_alpha/",
        file_path_template_map_fn=AzureUrl.AZURE_BLOB_STORAGE_HTTPS_URL_TEMPLATE.format,
    )
    assert my_data_connector.get_data_reference_count() == 4
    assert my_data_connector.get_data_references()[:3] == [
        "test_dir_alpha/A.csv",
        "test_dir_alpha/B.csv",
        "test_dir_alpha/C.csv",
    ]
    assert my_data_connector.get_matched_data_reference_count() == 4
    assert my_data_connector.get_matched_data_references()[:3] == [
        "test_dir_alpha/A.csv",
        "test_dir_alpha/B.csv",
        "test_dir_alpha/C.csv",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert my_data_connector.get_unmatched_data_reference_count() == 0

    my_batch_definition_list: List[LegacyBatchDefinition]
    my_batch_definition: LegacyBatchDefinition

    my_batch_request: BatchRequest

    # Try to fetch a batch from a nonexistent asset
    my_batch_request = BatchRequest(datasource_name="BASE", data_asset_name="A", options={})
    my_batch_definition_list = my_data_connector.get_batch_definition_list(
        batch_request=my_batch_request
    )
    assert len(my_batch_definition_list) == 0

    my_batch_request = BatchRequest(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_azure_blob_storage_data_asset",
        options={"part_1": "B"},
    )
    my_batch_definition_list = my_data_connector.get_batch_definition_list(
        batch_request=my_batch_request
    )
    assert len(my_batch_definition_list) == 1


@pytest.mark.big
@mock.patch(
    "great_expectations.datasource.fluent.data_asset.data_connector.azure_blob_storage_data_connector.list_azure_keys"
)
def test_foxtrot(mock_list_keys):
    mock_list_keys.return_value = []

    azure_client: azure.BlobServiceClient = cast("azure.BlobServiceClient", MockBlobServiceClient())

    my_data_connector: DataConnector

    my_data_connector = AzureBlobStorageDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_azure_blob_storage_data_asset",
        batching_regex=re.compile(r"(?P<part_1>.+)-(?P<part_2>.*)\.csv"),
        azure_client=azure_client,
        account_name="my_account",
        container="my_container",
        name_starts_with="",
        file_path_template_map_fn=AzureUrl.AZURE_BLOB_STORAGE_HTTPS_URL_TEMPLATE.format,
    )
    assert my_data_connector.get_data_reference_count() == 0
    assert my_data_connector.get_data_references()[:3] == []
    assert my_data_connector.get_matched_data_reference_count() == 0
    assert my_data_connector.get_matched_data_references()[:3] == []
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert my_data_connector.get_unmatched_data_reference_count() == 0

    mock_list_keys.return_value = [
        "test_dir_foxtrot/A/A-1.csv",
        "test_dir_foxtrot/A/A-2.csv",
        "test_dir_foxtrot/A/A-3.csv",
    ]

    my_data_connector = AzureBlobStorageDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_azure_blob_storage_data_asset",
        batching_regex=re.compile(r"(?P<part_1>.+)-(?P<part_2>.*)\.csv"),
        azure_client=azure_client,
        account_name="my_account",
        container="my_container",
        name_starts_with="test_dir_foxtrot/A/",
        file_path_template_map_fn=AzureUrl.AZURE_BLOB_STORAGE_HTTPS_URL_TEMPLATE.format,
    )

    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "test_dir_foxtrot/A/A-1.csv",
        "test_dir_foxtrot/A/A-2.csv",
        "test_dir_foxtrot/A/A-3.csv",
    ]
    assert my_data_connector.get_matched_data_reference_count() == 3
    assert my_data_connector.get_matched_data_references()[:3] == [
        "test_dir_foxtrot/A/A-1.csv",
        "test_dir_foxtrot/A/A-2.csv",
        "test_dir_foxtrot/A/A-3.csv",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert my_data_connector.get_unmatched_data_reference_count() == 0

    mock_list_keys.return_value = [
        "test_dir_foxtrot/B/B-1.txt",
        "test_dir_foxtrot/B/B-2.txt",
        "test_dir_foxtrot/B/B-3.txt",
    ]

    my_data_connector = AzureBlobStorageDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_azure_blob_storage_data_asset",
        batching_regex=re.compile(r"(?P<part_1>.+)-(?P<part_2>.*)\.txt"),
        azure_client=azure_client,
        account_name="my_account",
        container="my_container",
        name_starts_with="test_dir_foxtrot/B/",
        file_path_template_map_fn=AzureUrl.AZURE_BLOB_STORAGE_HTTPS_URL_TEMPLATE.format,
    )

    mock_list_keys.return_value = [
        "test_dir_foxtrot/B/B-1.txt",
        "test_dir_foxtrot/B/B-2.txt",
        "test_dir_foxtrot/B/B-3.txt",
    ]

    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "test_dir_foxtrot/B/B-1.txt",
        "test_dir_foxtrot/B/B-2.txt",
        "test_dir_foxtrot/B/B-3.txt",
    ]
    assert my_data_connector.get_matched_data_reference_count() == 3
    assert my_data_connector.get_matched_data_references()[:3] == [
        "test_dir_foxtrot/B/B-1.txt",
        "test_dir_foxtrot/B/B-2.txt",
        "test_dir_foxtrot/B/B-3.txt",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert my_data_connector.get_unmatched_data_reference_count() == 0

    my_data_connector = AzureBlobStorageDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_azure_blob_storage_data_asset",
        batching_regex=re.compile(r"(?P<part_1>.+)-(?P<part_2>.*)\.csv"),
        azure_client=azure_client,
        account_name="my_account",
        container="my_container",
        name_starts_with="test_dir_foxtrot/C/",
        file_path_template_map_fn=AzureUrl.AZURE_BLOB_STORAGE_HTTPS_URL_TEMPLATE.format,
    )

    mock_list_keys.return_value = [
        "test_dir_foxtrot/C/C-2017.csv",
        "test_dir_foxtrot/C/C-2018.csv",
        "test_dir_foxtrot/C/C-2019.csv",
    ]

    assert my_data_connector.get_data_reference_count() == 3
    assert my_data_connector.get_data_references()[:3] == [
        "test_dir_foxtrot/C/C-2017.csv",
        "test_dir_foxtrot/C/C-2018.csv",
        "test_dir_foxtrot/C/C-2019.csv",
    ]
    assert my_data_connector.get_matched_data_reference_count() == 3
    assert my_data_connector.get_matched_data_references()[:3] == [
        "test_dir_foxtrot/C/C-2017.csv",
        "test_dir_foxtrot/C/C-2018.csv",
        "test_dir_foxtrot/C/C-2019.csv",
    ]
    assert my_data_connector.get_unmatched_data_references()[:3] == []
    assert my_data_connector.get_unmatched_data_reference_count() == 0

    my_batch_request = BatchRequest(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_azure_blob_storage_data_asset",
        options={},
    )
    my_batch_definition_list: List[LegacyBatchDefinition] = (
        my_data_connector.get_batch_definition_list(batch_request=my_batch_request)
    )
    assert len(my_batch_definition_list) == 3


@pytest.mark.unit
@pytest.mark.parametrize(
    "whole_directory_override, expected_batch_count, expected_identifier_key",
    [
        pytest.param(
            True, 1, "path", id="with_whole_directory_override_returns_single_directory_batch"
        ),
        pytest.param(
            False,
            3,
            "filename",
            id="without_whole_directory_override_returns_individual_file_batches",
        ),
    ],
)
def test_abs_data_connector_whole_directory_path_override(
    whole_directory_override, expected_batch_count, expected_identifier_key, mocker
):
    """Test AzureBlobStorageDataConnector with whole_directory_path_override parameter."""
    # Setup
    account_name = "testaccount"
    container = "test-container"
    prefix = "test_directory/"
    whole_directory_path = f"wasbs://{container}@{account_name}.blob.core.windows.net/{prefix}"

    # Mock Azure client
    mock_azure_client = mocker.Mock()
    mock_container_client = mocker.Mock()
    mock_azure_client.get_container_client.return_value = mock_container_client

    # Create mock blobs for file mode testing
    mock_blobs = []
    test_files = [
        "test_directory/file1.csv",
        "test_directory/file2.csv",
        "test_directory/file3.csv",
    ]

    for file_path in test_files:
        mock_blob = mocker.Mock()
        mock_blob.name = file_path
        mock_blobs.append(mock_blob)

    mock_container_client.list_blobs.return_value = mock_blobs

    # Create data connector with conditional whole_directory_path_override
    data_connector = AzureBlobStorageDataConnector(
        datasource_name="my_abs_datasource",
        data_asset_name="my_data_asset",
        azure_client=mock_azure_client,
        account_name=account_name,
        container=container,
        name_starts_with=prefix,
        file_path_template_map_fn=lambda account_name,
        container,
        path: f"wasbs://{container}@{account_name}.blob.core.windows.net/{path}",
        whole_directory_path_override=whole_directory_path if whole_directory_override else None,
    )

    # Create batch request with conditional partitioner
    batch_request = BatchRequest(
        datasource_name="my_abs_datasource",
        data_asset_name="my_data_asset",
        options={},
        partitioner=None
        if whole_directory_override
        else FileNamePartitionerPath(regex=re.compile(r"(?P<filename>.+\.csv)")),
    )
    batch_definitions = data_connector.get_batch_definition_list(batch_request)

    # Verify expected batch count
    assert len(batch_definitions) == expected_batch_count

    # Verify batch definitions have correct structure
    for batch_definition in batch_definitions:
        assert batch_definition.datasource_name == "my_abs_datasource"
        assert batch_definition.data_asset_name == "my_data_asset"
        assert expected_identifier_key in batch_definition.batch_identifiers

    if whole_directory_override:
        # For directory mode, verify single batch with directory path
        batch_definition = batch_definitions[0]
        assert batch_definition.batch_identifiers["path"] == whole_directory_path
    else:
        # For file mode, verify individual file batches
        file_names = [bd.batch_identifiers["filename"] for bd in batch_definitions]
        expected_files = ["file1.csv", "file2.csv", "file3.csv"]
        assert sorted(file_names) == sorted(expected_files)


@pytest.mark.unit
def test_abs_data_connector_missing_file_path_template_map_fn_error():
    """Test Azure data connector raises error when file_path_template_map_fn is None."""
    azure_client: azure.BlobServiceClient = MockBlobServiceClient()

    data_connector = AzureBlobStorageDataConnector(
        datasource_name="my_abs_datasource",
        data_asset_name="my_data_asset",
        azure_client=azure_client,
        account_name="testaccount",
        container="test-container",
        name_starts_with="test/",
        file_path_template_map_fn=None,
    )

    with pytest.raises(MissingFilePathTemplateMapFnError):
        data_connector._get_full_file_path("test.csv")



from typing import ClassVar

from great_expectations.expectations.metrics.query_metrics.query_data_source_table.query_data_source_table import (  # noqa: E501  # too long, but rarely imported
    QueryDataSourceTable,
)


class ComparisonQueryDataSourceTable(QueryDataSourceTable):
    metric_name = "comparison_query.data_source_table"
    value_keys = ("comparison_query", "comparison_data_source_name")

    query_param_name: ClassVar[str] = "comparison_query"
    data_source_name_param_name: ClassVar[str] = "comparison_data_source_name"



import pytest

from great_expectations.expectations import (
    ExpectColumnMaxToBeBetween,
    ExpectColumnValuesToNotBeNull,
    UnexpectedRowsExpectation,
)
from great_expectations.expectations.window import Offset, Window


@pytest.mark.unit
@pytest.mark.parametrize(
    "expectation,expected_serialization",
    [
        (
            ExpectColumnMaxToBeBetween(
                column="test_column",
                min_value=1,
            ),
            (
                '{"id": null, "meta": null, "notes": null, "result_format": "BASIC", '
                '"description": null, "catch_exceptions": false, "rendered_content": null, '
                '"windows": null, "batch_id": null, "column": "test_column", '
                '"row_condition": null, "condition_parser": null, "min_value": 1.0, '
                '"max_value": null, "strict_min": false, "strict_max": false}'
            ),
        ),
        (
            ExpectColumnValuesToNotBeNull(
                column="test_column",
                mostly=0.82,
            ),
            (
                '{"id": null, "meta": null, "notes": null, "result_format": "BASIC", '
                '"description": null, "catch_exceptions": true, "rendered_content": null, '
                '"windows": null, "batch_id": null, "column": "test_column", "mostly": 0.82, '
                '"row_condition": null, "condition_parser": null}'
            ),
        ),
        (
            UnexpectedRowsExpectation(
                unexpected_rows_query="SELECT * FROM my_table WHERE data='bad'",
                description="Data shouldn't be bad.",
            ),
            (
                '{"id": null, "meta": null, "notes": null, "result_format": "BASIC", '
                '"description": "Data shouldn\'t be bad.", "catch_exceptions": false, '
                '"rendered_content": null, "windows": null, "batch_id": null, '
                '"unexpected_rows_query": "SELECT * FROM '
                "my_table WHERE data='bad'\"}"
            ),
        ),
        (
            ExpectColumnValuesToNotBeNull(
                column="test_column",
                mostly=0.82,
                windows=[
                    Window(
                        constraint_fn="a",
                        parameter_name="b",
                        range=5,
                        offset=Offset(positive=0.2, negative=0.2),
                    )
                ],
            ),
            (
                '{"id": null, "meta": null, "notes": null, "result_format": "BASIC", '
                '"description": null, "catch_exceptions": true, "rendered_content": null, '
                '"windows": [{"constraint_fn": "a", "parameter_name": "b", "range": 5, '
                '"offset": {"positive": 0.2, "negative": 0.2}, "strict": false}], '
                '"batch_id": null, "column": "test_column", "mostly": 0.82, '
                '"row_condition": null, "condition_parser": null}'
            ),
        ),
    ],
)
def test_expectation_serialization_snapshot(expectation, expected_serialization):
    assert expectation.json() == expected_serialization



from __future__ import annotations

from typing import Literal, Union

from great_expectations._docs_decorators import public_api
from great_expectations.compatibility.pydantic import PostgresDsn
from great_expectations.datasource.fluent.config_str import ConfigStr
from great_expectations.datasource.fluent.sql_datasource import SQLDatasource


@public_api
class PostgresDatasource(SQLDatasource):
    """Adds a postgres datasource to the data context.

    Args:
        name: The name of this postgres datasource.
        connection_string: The SQLAlchemy connection string used to connect to the postgres database.
            For example: "postgresql+psycopg2://postgres:@localhost/test_database"
        assets: An optional dictionary whose keys are TableAsset or QueryAsset names and whose values
            are TableAsset or QueryAsset objects.
    """  # noqa: E501 # FIXME CoP

    type: Literal["postgres"] = "postgres"  # type: ignore[assignment] # FIXME CoP
    connection_string: Union[ConfigStr, PostgresDsn]



from __future__ import annotations

import datetime
import json
import logging
import re
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Type,
    TypedDict,
    Union,
    overload,
)

import pandas as pd

from great_expectations._docs_decorators import deprecated_argument
from great_expectations.compatibility import pyspark
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.id_dict import BatchKwargs, BatchSpec, IDDict, IDDictID
from great_expectations.exceptions import InvalidBatchIdError
from great_expectations.types import DictDot, SerializableDictDot, safe_deep_copy
from great_expectations.util import (
    convert_to_json_serializable,  # noqa: TID251 # FIXME CoP
    deep_filter_properties_iterable,
    load_class,
)

if TYPE_CHECKING:
    from typing_extensions import NotRequired, TypeAlias

    from great_expectations.alias_types import JSONValues
    from great_expectations.datasource.fluent.data_connector.batch_filter import BatchSlice
    from great_expectations.datasource.fluent.interfaces import (
        Batch as FluentBatch,
    )
    from great_expectations.datasource.fluent.interfaces import (
        BatchParameters,
    )
    from great_expectations.datasource.fluent.interfaces import (
        BatchRequest as FluentBatchRequest,
    )
    from great_expectations.validator.metrics_calculator import MetricsCalculator


logger = logging.getLogger(__name__)


class BlockConfigBatchRequestTypedDict(TypedDict):
    datasource_name: str
    data_connector_name: str
    data_asset_name: str
    runtime_parameters: NotRequired[dict]
    batch_identifiers: NotRequired[dict]
    batch_spec_passthrough: NotRequired[dict]
    data_connector_query: NotRequired[dict]
    limit: NotRequired[BatchSlice]


def _get_fluent_batch_request_class() -> Type[FluentBatchRequest]:
    """Using this function helps work around circular import dependncies."""
    module_name = "great_expectations.datasource.fluent.batch_request"
    class_name = "BatchRequest"
    return load_class(class_name=class_name, module_name=module_name)


def _get_fluent_batch_class() -> Type[FluentBatch]:
    """Using this function helps work around circular import dependncies."""
    module_name = "great_expectations.datasource.fluent.interfaces"
    class_name = "Batch"
    return load_class(class_name=class_name, module_name=module_name)


def _get_metrics_calculator_class() -> Type[MetricsCalculator]:
    """Using this function helps work around circular import dependncies."""
    module_name = "great_expectations.validator.metrics_calculator"
    class_name = "MetricsCalculator"
    return load_class(class_name=class_name, module_name=module_name)


class LegacyBatchDefinition(SerializableDictDot):
    """Precisely identifies a set of data from a data source.

    More concretely, a BatchDefinition includes all the information required to precisely
    identify a set of data from the external data source that should be
    translated into a Batch. One or more BatchDefinitions should always be
    *returned* from the Datasource, as a result of processing the Batch Request.

    ---Documentation---
            - https://docs.greatexpectations.io/docs/terms/batch/#batches-and-batch-requests-design-motivation

    Args:
        datasource_name: name of the Datasource used to connect to the data
        data_connector_name: name of the DataConnector used to connect to the data
        data_asset_name: name of the DataAsset used to connect to the data
        batch_identifiers: key-value pairs that the DataConnector
            will use to obtain a specific set of data
        batch_spec_passthrough: a dictionary of additional parameters that
            the ExecutionEngine will use to obtain a specific set of data

    Returns:
        BatchDefinition
    """

    def __init__(  # noqa: PLR0913 # FIXME CoP
        self,
        datasource_name: str,
        data_connector_name: str,
        data_asset_name: str,
        batch_identifiers: IDDict,
        batch_spec_passthrough: dict | None = None,
        batching_regex: re.Pattern | None = None,
    ) -> None:
        self._validate_batch_definition(
            datasource_name=datasource_name,
            data_connector_name=data_connector_name,
            data_asset_name=data_asset_name,
            batch_identifiers=batch_identifiers,
        )

        assert type(batch_identifiers) == IDDict  # noqa: E721 # legacy code

        self._datasource_name = datasource_name
        self._data_connector_name = data_connector_name
        self._data_asset_name = data_asset_name
        self._batch_identifiers = batch_identifiers
        self._batch_spec_passthrough = batch_spec_passthrough
        self._batching_regex = batching_regex

    @override
    def to_json_dict(self) -> dict[str, JSONValues]:
        """Returns a JSON-serializable dict representation of this BatchDefinition.

        Returns:
            A JSON-serializable dict representation of this BatchDefinition.
        """
        fields_dict: dict = {
            "datasource_name": self._datasource_name,
            "data_connector_name": self._data_connector_name,
            "data_asset_name": self._data_asset_name,
            "batch_identifiers": self._batch_identifiers,
        }
        if self._batch_spec_passthrough:
            fields_dict["batch_spec_passthrough"] = self._batch_spec_passthrough
        if self._batching_regex:
            fields_dict["batching_regex"] = self._batching_regex

        return convert_to_json_serializable(data=fields_dict)

    @override
    def __repr__(self) -> str:
        doc_fields_dict: dict = {
            "datasource_name": self._datasource_name,
            "data_connector_name": self._data_connector_name,
            "data_asset_name": self._data_asset_name,
            "batch_identifiers": self._batch_identifiers,
        }
        return str(doc_fields_dict)

    @staticmethod
    def _validate_batch_definition(
        datasource_name: str,
        data_connector_name: str,
        data_asset_name: str,
        batch_identifiers: IDDict,
    ) -> None:
        if datasource_name is None:
            raise ValueError("A valid datasource must be specified.")  # noqa: TRY003 # FIXME CoP
        if datasource_name and not isinstance(datasource_name, str):
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                f"""The type of an datasource name must be a string (Python "str").  The type given is
"{type(datasource_name)!s}", which is illegal.
            """  # noqa: E501 # FIXME CoP
            )
        if data_connector_name is None:
            raise ValueError("A valid data_connector must be specified.")  # noqa: TRY003 # FIXME CoP
        if data_connector_name and not isinstance(data_connector_name, str):
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                f"""The type of a data_connector name must be a string (Python "str").  The type given is
"{type(data_connector_name)!s}", which is illegal.
                """  # noqa: E501 # FIXME CoP
            )
        if data_asset_name is None:
            raise ValueError("A valid data_asset_name must be specified.")  # noqa: TRY003 # FIXME CoP
        if data_asset_name and not isinstance(data_asset_name, str):
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                f"""The type of a data_asset name must be a string (Python "str").  The type given is
"{type(data_asset_name)!s}", which is illegal.
                """  # noqa: E501 # FIXME CoP
            )
        if batch_identifiers and not isinstance(batch_identifiers, IDDict):
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                f"""The type of batch_identifiers must be an IDDict object.  The type given is \
"{type(batch_identifiers)!s}", which is illegal.
"""
            )

    @property
    def datasource_name(self) -> str:
        return self._datasource_name

    @property
    def data_connector_name(self) -> str:
        return self._data_connector_name

    @property
    def data_asset_name(self) -> str:
        return self._data_asset_name

    @property
    def batch_identifiers(self) -> IDDict:
        return self._batch_identifiers

    @property
    def batch_spec_passthrough(self) -> dict | None:
        return self._batch_spec_passthrough

    @batch_spec_passthrough.setter
    def batch_spec_passthrough(self, batch_spec_passthrough: dict | None) -> None:
        self._batch_spec_passthrough = batch_spec_passthrough

    @property
    def id(self) -> IDDictID:
        return IDDict(self.to_json_dict()).to_id()

    @property
    def batching_regex(self) -> re.Pattern | None:
        return self._batching_regex

    @override
    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            # Delegate comparison to the other instance's __eq__.
            return NotImplemented
        return self.id == other.id

    @override
    def __str__(self):
        return json.dumps(self.to_json_dict(), indent=2)

    @override
    def __hash__(self) -> int:
        """Overrides the default implementation"""
        _result_hash: int = hash(self.id)
        return _result_hash


class BatchRequestBase(SerializableDictDot):
    """
    This class is for internal inter-object protocol purposes only.
    As such, it contains all attributes of a batch_request, but does not validate them.
    See the BatchRequest class, which extends BatchRequestBase and validates the attributes.

    BatchRequestBase is used for the internal protocol purposes exclusively, not part of API for the developer users.

    Previously, the very same BatchRequest was used for both the internal protocol purposes and as part of the API
    exposed to developers.  However, while convenient for internal data interchange, using the same BatchRequest class
    as arguments to the externally-exported DataContext.get_batch_list() and DataContext.get_validator() API calls for
    obtaining batches and/or validators was insufficiently expressive to fulfill the needs of both. In the user-accessible
    API, BatchRequest, must enforce that all members of the triple, consisting of data_source_name, data_connector_name,
    and data_asset_name, are not NULL.  Whereas for the internal protocol, BatchRequest is used as a flexible bag of attributes,
    in which any fields are allowed to be NULL.  Hence, now, BatchRequestBase is dedicated for the use as the bag oof attributes
    for the internal protocol use, whereby NULL values are allowed as per the internal needs.  The BatchRequest class extends
    BatchRequestBase and adds to it strong validation (described above plus additional attribute validation) so as to formally
    validate user specified fields.
    """  # noqa: E501 # FIXME CoP

    def __init__(  # noqa: PLR0913 # FIXME CoP
        self,
        datasource_name: str,
        data_connector_name: str,
        data_asset_name: str,
        data_connector_query: dict | None = None,
        limit: int | None = None,
        runtime_parameters: dict | None = None,
        batch_identifiers: dict | None = None,
        batch_spec_passthrough: dict | None = None,
    ) -> None:
        self._datasource_name = datasource_name
        self._data_connector_name = data_connector_name
        self._data_asset_name = data_asset_name
        self._data_connector_query = data_connector_query
        self._limit = limit

        self._runtime_parameters = runtime_parameters
        self._batch_identifiers = batch_identifiers
        self._batch_spec_passthrough = batch_spec_passthrough

    @property
    def datasource_name(self) -> str:
        return self._datasource_name

    @datasource_name.setter
    def datasource_name(self, value: str) -> None:
        self._datasource_name = value

    @property
    def data_connector_name(self) -> str:
        return self._data_connector_name

    @data_connector_name.setter
    def data_connector_name(self, value: str) -> None:
        self._data_connector_name = value

    @property
    def data_asset_name(self) -> str:
        return self._data_asset_name

    @data_asset_name.setter
    def data_asset_name(self, data_asset_name) -> None:
        self._data_asset_name = data_asset_name

    @property
    def data_connector_query(self) -> dict | None:
        return self._data_connector_query

    @data_connector_query.setter
    def data_connector_query(self, value: dict) -> None:
        self._data_connector_query = value

    @property
    def limit(self) -> int | None:
        return self._limit

    @limit.setter
    def limit(self, value: int) -> None:
        self._limit = value

    @property
    def runtime_parameters(self) -> dict | None:
        return self._runtime_parameters

    @runtime_parameters.setter
    def runtime_parameters(self, value: dict) -> None:
        self._runtime_parameters = value

    @property
    def batch_identifiers(self) -> dict | None:
        return self._batch_identifiers

    @batch_identifiers.setter
    def batch_identifiers(self, value: dict) -> None:
        self._batch_identifiers = value

    @property
    def batch_spec_passthrough(self) -> dict | None:
        return self._batch_spec_passthrough

    @batch_spec_passthrough.setter
    def batch_spec_passthrough(self, value: dict) -> None:
        self._batch_spec_passthrough = value

    @property
    def id(self) -> IDDictID:
        return IDDict(self.to_json_dict()).to_id()

    @override
    def to_dict(self) -> BlockConfigBatchRequestTypedDict:  # type: ignore[override] # TypedDict is more specific dict type
        return standardize_batch_request_display_ordering(
            batch_request=super().to_dict()  # type: ignore[arg-type] # TypedDict is more specific dict type
        )

    # While this class is private, it is inherited from and this method is part
    # of the public api on the child.
    @override
    def to_json_dict(self) -> dict[str, JSONValues]:
        """Returns a JSON-serializable dict representation of this BatchRequestBase.

        Returns:
            A JSON-serializable dict representation of this BatchRequestBase.
        """
        # TODO: <Alex>2/4/2022</Alex>
        # This implementation of "SerializableDictDot.to_json_dict() occurs frequently and should ideally serve as the  # noqa: E501 # FIXME CoP
        # reference implementation in the "SerializableDictDot" class itself.  However, the circular import dependencies,  # noqa: E501 # FIXME CoP
        # due to the location of the "great_expectations/types/__init__.py" and "great_expectations/core/util.py" modules  # noqa: E501 # FIXME CoP
        # make this refactoring infeasible at the present time.

        # if batch_data appears in BatchRequest, temporarily replace it with
        # str placeholder before calling convert_to_json_serializable so that
        # batch_data is not serialized
        serializeable_dict: dict
        if batch_request_contains_batch_data(batch_request=self):
            if self.runtime_parameters is None:
                raise ValueError("BatchRequestBase missing runtime_parameters during serialization")  # noqa: TRY003 # FIXME CoP
            batch_data: BatchRequestBase | dict = self.runtime_parameters["batch_data"]
            self.runtime_parameters["batch_data"] = str(type(batch_data))

            serializeable_dict = convert_to_json_serializable(data=self.to_dict())  # type: ignore[call-overload] # TypedDict is more specific dict type
            # after getting serializable_dict, restore original batch_data
            self.runtime_parameters["batch_data"] = batch_data
        else:
            serializeable_dict = convert_to_json_serializable(data=self.to_dict())  # type: ignore[call-overload] # TypedDict is more specific dict type

        return serializeable_dict

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)

        memo[id(self)] = result

        for key, value in self.to_raw_dict().items():
            value_copy = safe_deep_copy(data=value, memo=memo)
            setattr(result, key, value_copy)

        return result

    @override
    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            # Delegate comparison to the other instance's __eq__.
            return NotImplemented

        return self.id == other.id

    @override
    def __hash__(self) -> int:
        return hash(self.id)

    @override
    def __repr__(self) -> str:
        """
        # TODO: <Alex>2/4/2022</Alex>
        This implementation of a custom "__repr__()" occurs frequently and should ideally serve as the reference
        implementation in the "SerializableDictDot" class.  However, the circular import dependencies, due to the
        location of the "great_expectations/types/__init__.py" and "great_expectations/core/util.py" modules make this
        refactoring infeasible at the present time.
        """  # noqa: E501 # FIXME CoP
        json_dict: dict = self.to_json_dict()
        deep_filter_properties_iterable(
            properties=json_dict,
            inplace=True,
        )
        return json.dumps(json_dict, indent=2)

    @override
    def __str__(self) -> str:
        """
        # TODO: <Alex>2/4/2022</Alex>
        This implementation of a custom "__str__()" occurs frequently and should ideally serve as the reference
        implementation in the "SerializableDictDot" class.  However, the circular import dependencies, due to the
        location of the "great_expectations/types/__init__.py" and "great_expectations/core/util.py" modules make this
        refactoring infeasible at the present time.
        """  # noqa: E501 # FIXME CoP
        return self.__repr__()

    @staticmethod
    def _validate_init_parameters(
        datasource_name: str,
        data_connector_name: str,
        data_asset_name: str,
        data_connector_query: dict | None = None,
        limit: int | None = None,
    ) -> None:
        # TODO test and check all logic in this validator!
        if not (datasource_name and isinstance(datasource_name, str)):
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                f"""The type of an datasource name must be a string (Python "str").  The type given is
"{type(datasource_name)!s}", which is illegal.
            """  # noqa: E501 # FIXME CoP
            )
        if not (data_connector_name and isinstance(data_connector_name, str)):
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                f"""The type of data_connector name must be a string (Python "str").  The type given is
"{type(data_connector_name)!s}", which is illegal.
                """  # noqa: E501 # FIXME CoP
            )
        if not (data_asset_name and isinstance(data_asset_name, str)):
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                f"""The type of data_asset name must be a string (Python "str").  The type given is
        "{type(data_asset_name)!s}", which is illegal.
                        """
            )
        # TODO Abe 20201015: Switch this to DataConnectorQuery.
        if data_connector_query and not isinstance(data_connector_query, dict):
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                f"""The type of data_connector_query must be a dict object.  The type given is
"{type(data_connector_query)!s}", which is illegal.
                """
            )
        if limit and not isinstance(limit, int):
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                f"""The type of limit must be an integer (Python "int").  The type given is "{type(limit)!s}", which
is illegal.
                """  # noqa: E501 # FIXME CoP
            )


class BatchRequest(BatchRequestBase):
    """A BatchRequest is the way to specify which data Great Expectations will validate.

    A Batch Request is provided to a Datasource in order to create a Batch.

    ---Documentation---
        - https://docs.greatexpectations.io/docs/guides/connecting_to_your_data/how_to_get_one_or_more_batches_of_data_from_a_configured_datasource/#1-construct-a-batchrequest
        - https://docs.greatexpectations.io/docs/terms/batch_request

    The `data_connector_query` parameter can include an index slice:

    ```python
    {
        "index": "-3:"
    }
    ```

    or it can include a filter:

    ```python
    {
        "batch_filter_parameters": {"year": "2020"}
    }
    ```

    Args:
        datasource_name: name of the Datasource used to connect to the data
        data_connector_name: name of the DataConnector used to connect to the data
        data_asset_name: name of the DataAsset used to connect to the data
        data_connector_query: a dictionary of query parameters the DataConnector
            should use to filter the batches returned from a BatchRequest
        limit: if specified, the maximum number of *batches* to be returned
            (limit does not affect the number of records in each batch)
        batch_spec_passthrough: a dictionary of additional parameters that
            the ExecutionEngine will use to obtain a specific set of data

    Returns:
        BatchRequest
    """

    include_field_names: ClassVar[set[str]] = {
        "datasource_name",
        "data_connector_name",
        "data_asset_name",
        "data_connector_query",
        "limit",
        "batch_spec_passthrough",
    }

    def __init__(  # noqa: PLR0913 # FIXME CoP
        self,
        datasource_name: str,
        data_connector_name: str,
        data_asset_name: str,
        data_connector_query: dict | None = None,
        limit: int | None = None,
        batch_spec_passthrough: dict | None = None,
    ) -> None:
        self._validate_init_parameters(
            datasource_name=datasource_name,
            data_connector_name=data_connector_name,
            data_asset_name=data_asset_name,
            data_connector_query=data_connector_query,
            limit=limit,
        )
        super().__init__(
            datasource_name=datasource_name,
            data_connector_name=data_connector_name,
            data_asset_name=data_asset_name,
            data_connector_query=data_connector_query,
            limit=limit,
            batch_spec_passthrough=batch_spec_passthrough,
        )


class RuntimeBatchRequest(BatchRequestBase):
    """A RuntimeBatchRequest creates a Batch for a RuntimeDataConnector.

    Instead of serving as a description of what data Great Expectations should
    fetch, a RuntimeBatchRequest serves as a wrapper for data that is passed in
    at runtime (as an in-memory dataframe, file/S3 path, or SQL query), with
    user-provided identifiers for uniquely identifying the data.

    ---Documentation---
        - https://docs.greatexpectations.io/docs/terms/batch_request/#runtimedataconnector-and-runtimebatchrequest
        - https://docs.greatexpectations.io/docs/guides/connecting_to_your_data/how_to_configure_a_runtimedataconnector/

    runtime_parameters will vary depending on the Datasource used with the data.

    For a dataframe:

    ```python
    {"batch_data": df}
    ```

    For a path on a filesystem:

    ```python
        {"path": "/path/to/data/file.csv"}
    ```

    Args:
        datasource_name: name of the Datasource used to connect to the data
        data_connector_name: name of the DataConnector used to connect to the data
        data_asset_name: name of the DataAsset used to connect to the data
        runtime_parameters: a dictionary containing the data to process,
            a path to the data, or a query, depending on the associated Datasource
        batch_identifiers: a dictionary to serve as a persistent, unique
            identifier for the data included in the Batch
        batch_spec_passthrough: a dictionary of additional parameters that
            the ExecutionEngine will use to obtain a specific set of data
    Returns:
        BatchRequest
    """

    include_field_names: ClassVar[set[str]] = {
        "datasource_name",
        "data_connector_name",
        "data_asset_name",
        "runtime_parameters",
        "batch_identifiers",
        "batch_spec_passthrough",
    }

    def __init__(  # noqa: PLR0913 # FIXME CoP
        self,
        datasource_name: str,
        data_connector_name: str,
        data_asset_name: str,
        runtime_parameters: dict,
        batch_identifiers: dict,
        batch_spec_passthrough: dict | None = None,
    ) -> None:
        self._validate_init_parameters(
            datasource_name=datasource_name,
            data_connector_name=data_connector_name,
            data_asset_name=data_asset_name,
        )
        self._validate_runtime_batch_request_specific_init_parameters(
            runtime_parameters=runtime_parameters,
            batch_identifiers=batch_identifiers,
            batch_spec_passthrough=batch_spec_passthrough,
        )
        super().__init__(
            datasource_name=datasource_name,
            data_connector_name=data_connector_name,
            data_asset_name=data_asset_name,
            runtime_parameters=runtime_parameters,
            batch_identifiers=batch_identifiers,
            batch_spec_passthrough=batch_spec_passthrough,
        )

    @staticmethod
    def _validate_runtime_batch_request_specific_init_parameters(
        runtime_parameters: dict | None,
        batch_identifiers: dict | None,
        batch_spec_passthrough: dict | None = None,
    ) -> None:
        """
        We must have both or neither of runtime_parameters and batch_identifiers (but not either one of them).
        This is strict equivalence ("if-and-only") condition ("exclusive NOR"); otherwise, ("exclusive OR") means error.
        """  # noqa: E501 # FIXME CoP
        if (not runtime_parameters and batch_identifiers) or (
            runtime_parameters and not batch_identifiers
        ):
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                "It must be that either both runtime_parameters and batch_identifiers are present, or both are missing"  # noqa: E501 # FIXME CoP
            )

        # if there is a value, make sure it is a dict
        if runtime_parameters and not (isinstance(runtime_parameters, dict)):
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                f"""The runtime_parameters must be a non-empty dict object.
                The type given is "{type(runtime_parameters)!s}", which is an illegal type or an empty dictionary."""  # noqa: E501 # FIXME CoP
            )

        # if there is a value, make sure it is a dict
        if batch_identifiers and not isinstance(batch_identifiers, dict):
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                f"""The type for batch_identifiers must be a dict object, with keys being identifiers defined in the
                data connector configuration.  The type given is "{type(batch_identifiers)!s}", which is illegal."""  # noqa: E501 # FIXME CoP
            )

        if batch_spec_passthrough and not (isinstance(batch_spec_passthrough, dict)):
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                f"""The type for batch_spec_passthrough must be a dict object. The type given is \
"{type(batch_spec_passthrough)!s}", which is illegal.
"""
            )


# TODO: <Alex>The following class is to support the backward compatibility with the legacy design.</Alex>  # noqa: E501 # FIXME CoP
class BatchMarkers(BatchKwargs):
    """A BatchMarkers is a special type of BatchKwargs (so that it has a batch_fingerprint) but it generally does
    NOT require specific keys and instead captures information about the OUTPUT of a datasource's fetch
    process, such as the timestamp at which a query was executed."""  # noqa: E501 # FIXME CoP

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if "ge_load_time" not in self:
            raise InvalidBatchIdError("BatchMarkers requires a ge_load_time")  # noqa: TRY003 # FIXME CoP

    @property
    def ge_load_time(self):
        return self.get("ge_load_time")


class BatchData:
    def __init__(self, execution_engine) -> None:
        self._execution_engine = execution_engine

    @property
    def execution_engine(self):
        return self._execution_engine

    # noinspection PyMethodMayBeStatic
    def head(self, *args, **kwargs) -> pd.DataFrame:
        # CONFLICT ON PURPOSE. REMOVE.
        return pd.DataFrame({})


# TODO: <Alex>This module needs to be cleaned up.
#  We have Batch used for the legacy design, and we also need Batch for the new design.
#  However, right now, the Batch from the legacy design is imported into execution engines of the new design.  # noqa: E501 # FIXME CoP
#  As a result, we have multiple, inconsistent versions of BatchMarkers, extending legacy/new classes.</Alex>  # noqa: E501 # FIXME CoP
# TODO: <Alex>See also "great_expectations/datasource/types/batch_spec.py".</Alex>
@deprecated_argument(argument_name="data_context", version="0.14.0")
@deprecated_argument(argument_name="datasource_name", version="0.14.0")
@deprecated_argument(argument_name="batch_parameters", version="0.14.0")
@deprecated_argument(argument_name="batch_kwargs", version="0.14.0")
class Batch(SerializableDictDot):
    """A Batch is a selection of records from a Data Asset.

    A Datasource produces Batch objects to interact directly with data. Creating
    a Batch does NOT require moving data; the Batch facilitates access to the
    data and maintains metadata.

    ---Documentation---
            - https://docs.greatexpectations.io/docs/terms/batch/

    Args:
        data: A BatchDataType object which interacts directly with the
            ExecutionEngine.
        batch_request: BatchRequest that was used to obtain the data.
        batch_definition: Complete BatchDefinition that describes the data.
        batch_spec: Complete BatchSpec that describes the data.
        batch_markers: Additional metadata that may be useful to understand
            batch.
        data_context: DataContext connected to the
        datasource_name: name of datasource used to obtain the batch
        batch_parameters: keyword arguments describing the batch data
        batch_kwargs: keyword arguments used to request a batch from a Datasource

    Returns:
        Batch instance created.
    """

    def __init__(  # noqa: PLR0913 # FIXME CoP
        self,
        data: BatchDataType | None = None,
        batch_request: BatchRequestBase | dict | None = None,
        batch_definition: LegacyBatchDefinition | None = None,
        batch_spec: BatchSpec | None = None,
        batch_markers: BatchMarkers | None = None,
        # The remaining parameters are for backward compatibility.
        data_context=None,
        datasource_name=None,
        batch_parameters=None,
        batch_kwargs=None,
    ) -> None:
        self._data = data
        if batch_request is None:
            batch_request = {}

        self._batch_request = batch_request

        if batch_definition is None:
            batch_definition = IDDict()  # type: ignore[assignment] # FIXME CoP

        self._batch_definition = batch_definition

        if batch_spec is None:
            batch_spec = BatchSpec()

        self._batch_spec = batch_spec

        if batch_markers is None:
            batch_markers = BatchMarkers(
                {
                    "ge_load_time": datetime.datetime.now(datetime.timezone.utc).strftime(
                        "%Y%m%dT%H%M%S.%fZ"
                    )
                }
            )

        self._batch_markers = batch_markers

        # The remaining parameters are for backward compatibility.
        self._data_context = data_context
        self._datasource_name = datasource_name
        self._batch_parameters = batch_parameters
        self._batch_kwargs = batch_kwargs or BatchKwargs()

    @property
    def data(self) -> BatchDataType | None:
        """Getter for Batch data"""
        return self._data

    @data.setter
    def data(self, value: BatchDataType) -> None:
        """Setter for Batch data"""
        self._data = value

    @property
    def batch_request(self):
        return self._batch_request

    @batch_request.setter
    def batch_request(self, batch_request) -> None:
        self._batch_request = batch_request

    @property
    def batch_definition(self):
        return self._batch_definition

    @batch_definition.setter
    def batch_definition(self, batch_definition) -> None:
        self._batch_definition = batch_definition

    @property
    def batch_spec(self):
        return self._batch_spec

    @property
    def batch_markers(self):
        return self._batch_markers

    # The remaining properties are for backward compatibility.
    @property
    def data_context(self):
        return self._data_context

    @property
    def datasource_name(self):
        return self._datasource_name

    @property
    def batch_parameters(self):
        return self._batch_parameters

    @property
    def batch_kwargs(self):
        return self._batch_kwargs

    @override
    def to_dict(self) -> dict:
        dict_obj: dict = {
            "data": str(self.data),
            "batch_request": self.batch_request.to_dict(),
            "batch_definition": self.batch_definition.to_json_dict()
            if isinstance(self.batch_definition, LegacyBatchDefinition)
            else {},
            "batch_spec": self.batch_spec,
            "batch_markers": self.batch_markers,
        }
        return dict_obj

    @override
    def to_json_dict(self) -> dict[str, JSONValues]:
        """Returns a JSON-serializable dict representation of this Batch.

        Returns:
            A JSON-serializable dict representation of this Batch.
        """
        json_dict: dict = self.to_dict()
        deep_filter_properties_iterable(
            properties=json_dict["batch_request"],
            inplace=True,
        )
        return json_dict

    @property
    def id(self):
        batch_definition = self._batch_definition
        if isinstance(batch_definition, LegacyBatchDefinition):
            return batch_definition.id

        if isinstance(batch_definition, IDDict):
            return batch_definition.to_id()

        if isinstance(batch_definition, dict):
            return IDDict(batch_definition).to_id()

        return IDDict({}).to_id()

    @override
    def __str__(self):
        return json.dumps(self.to_json_dict(), indent=2)

    def head(self, n_rows: int = 5, fetch_all: bool = False) -> pd.DataFrame:
        """Return the first n rows from the Batch.

        This function returns the first n_rows rows. It is useful for quickly testing
        if your object has the data you expected.

        It will always obtain data from the Datasource and return a Pandas
        DataFrame available locally.

        Args:
             n_rows: the number of rows to return
             fetch_all: whether to fetch all rows; overrides n_rows if set to True

        Returns:
            A Pandas DataFrame
        """
        self._data.execution_engine.batch_manager.load_batch_list(batch_list=[self])  # type: ignore[union-attr] # FIXME CoP
        metrics_calculator = _get_metrics_calculator_class()(
            execution_engine=self._data.execution_engine,  # type: ignore[union-attr, arg-type] # FIXME CoP
            show_progress_bars=True,
        )
        table_head_df: pd.DataFrame = metrics_calculator.head(
            n_rows=n_rows,
            domain_kwargs={"batch_id": self.id},
            fetch_all=fetch_all,
        )
        return table_head_df


def materialize_batch_request(
    batch_request: BatchRequestBase | dict | None = None,
) -> FluentBatchRequest | BatchRequestBase | None:
    def _is_fluent_batch_request(
        args: dict[str, Any] | BlockConfigBatchRequestTypedDict,
    ) -> bool:
        from great_expectations.datasource.fluent.constants import _DATA_CONNECTOR_NAME

        return (
            ("datasource_name" in args)
            and ("data_asset_name" in args)
            and (
                ("data_connector_name" not in args)
                or (args["data_connector_name"].lower() == _DATA_CONNECTOR_NAME)
            )
        )

    effective_batch_request = get_batch_request_as_dict(batch_request=batch_request)

    if not effective_batch_request:
        return None

    batch_request_class: type
    if _is_fluent_batch_request(args=effective_batch_request):
        batch_request_class = _get_fluent_batch_request_class()
    elif batch_request_contains_runtime_parameters(batch_request=effective_batch_request):
        batch_request_class = RuntimeBatchRequest
    else:
        batch_request_class = BatchRequest

    return batch_request_class(**effective_batch_request)


def batch_request_contains_batch_data(
    batch_request: BatchRequestBase
    | FluentBatchRequest
    | dict
    | BlockConfigBatchRequestTypedDict
    | None = None,
) -> bool:
    return (
        batch_request_contains_runtime_parameters(batch_request=batch_request)
        and batch_request["runtime_parameters"].get("batch_data") is not None  # type: ignore[index] # FDS BatchRequest is pydantic model
    )


def batch_request_contains_runtime_parameters(
    batch_request: BatchRequestBase
    | FluentBatchRequest
    | dict
    | BlockConfigBatchRequestTypedDict
    | None = None,
) -> bool:
    return (
        batch_request is not None
        and isinstance(batch_request, (dict, DictDot))
        and batch_request.get("runtime_parameters") is not None
    )


@overload
def get_batch_request_as_dict(
    batch_request: BatchRequestBase
    | FluentBatchRequest
    | dict
    | BlockConfigBatchRequestTypedDict = ...,
) -> dict: ...


@overload
def get_batch_request_as_dict(
    batch_request: None = ...,
) -> None: ...


def get_batch_request_as_dict(
    batch_request: BatchRequestBase
    | FluentBatchRequest
    | dict
    | BlockConfigBatchRequestTypedDict
    | None = None,
) -> dict | None:
    if batch_request is None:
        return None

    if isinstance(batch_request, (BatchRequest, RuntimeBatchRequest)):
        batch_request = batch_request.to_dict()

    if isinstance(batch_request, _get_fluent_batch_request_class()):
        batch_request = batch_request.dict()

    return batch_request  # type: ignore[return-value] # FDS BatchRequest is missing data_connector (other fields)


def _get_block_batch_request(  # noqa: C901, PLR0913 # FIXME CoP
    *,
    datasource_name: str | None = None,
    data_connector_name: str | None = None,
    data_asset_name: str | None = None,
    batch_identifiers: dict | None = None,
    batch_spec_passthrough: dict | None = None,
    batch_filter_parameters: dict | None = None,
    data_connector_query: dict | None = None,
    limit: int | None = None,
    index: int | list | tuple | slice | str | None = None,
    custom_filter_function: Callable | None = None,
    sampling_method: str | None = None,
    sampling_kwargs: dict | None = None,
    partitioner_method: str | None = None,
    partitioner_kwargs: dict | None = None,
    **kwargs,
):
    """Returns a block-config batch request based on the provided parameters

    raises:
        ValueError if the parameters don't support creating a block-config batch request
    """
    if data_connector_query is None:
        if batch_filter_parameters is not None and batch_identifiers is not None:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                'Must provide either "batch_filter_parameters" or "batch_identifiers", not both.'
            )

        if batch_filter_parameters is None and batch_identifiers is not None:
            logger.warning(
                'Attempting to build data_connector_query but "batch_identifiers" was provided '
                'instead of "batch_filter_parameters". The "batch_identifiers" key on '
                'data_connector_query has been renamed to "batch_filter_parameters". Please update '
                'your code. Falling back on provided "batch_identifiers".'
            )
            batch_filter_parameters = batch_identifiers
        elif batch_filter_parameters is None and batch_identifiers is None:
            batch_filter_parameters = kwargs
        else:
            # Raise a warning if kwargs exist
            pass

        data_connector_query_params: dict = {
            "batch_filter_parameters": batch_filter_parameters,
            "limit": limit,
            "index": index,
            "custom_filter_function": custom_filter_function,
        }
        data_connector_query = IDDict(data_connector_query_params)
    else:
        # TODO: Raise a warning if batch_filter_parameters or kwargs exist
        data_connector_query = IDDict(data_connector_query)

    if batch_spec_passthrough is None:
        batch_spec_passthrough = {}
        if sampling_method is not None:
            sampling_params: dict = {
                "sampling_method": sampling_method,
            }
            if sampling_kwargs is not None:
                sampling_params["sampling_kwargs"] = sampling_kwargs
            batch_spec_passthrough.update(sampling_params)
        if partitioner_method is not None:
            partitioner_params: dict = {
                "partitioner_method": partitioner_method,
            }
            if partitioner_kwargs is not None:
                partitioner_params["partitioner_kwargs"] = partitioner_kwargs
            batch_spec_passthrough.update(partitioner_params)

    batch_request_as_dict: dict = {
        "datasource_name": datasource_name,
        "data_connector_name": data_connector_name,
        "data_asset_name": data_asset_name,
        "data_connector_query": data_connector_query,
        "batch_spec_passthrough": batch_spec_passthrough,
    }

    deep_filter_properties_iterable(
        properties=batch_request_as_dict,
        inplace=True,
    )

    return BatchRequest(**batch_request_as_dict)


def _get_runtime_batch_request(  # noqa: PLR0913 # FIXME CoP
    *,
    datasource_name: str,
    data_connector_name: str,
    data_asset_name: str,
    runtime_parameters: dict | None = None,
    batch_identifiers: dict | None = None,
    batch_spec_passthrough: dict | None = None,
    batch_data: Any | None = None,
    query: str | None = None,
    path: str | None = None,
    **kwargs,
) -> RuntimeBatchRequest | None:
    """Returns a `RuntimeBatchRequest`, or `None` if the arguments don't support it."""
    if any(
        [batch_data, query, path, runtime_parameters]
    ):  # one of these must be specified for runtime batch requests
        # parameter checking
        if len([arg for arg in [batch_data, query, path] if arg is not None]) > 1:
            raise ValueError("Must provide only one of batch_data, query, or path.")  # noqa: TRY003 # FIXME CoP

        if runtime_parameters and any(
            [
                batch_data and "batch_data" in runtime_parameters,
                query and "query" in runtime_parameters,
                path and "path" in runtime_parameters,
            ]
        ):
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                "If batch_data, query, or path arguments are provided, the same keys cannot appear in the "  # noqa: E501 # FIXME CoP
                "runtime_parameters argument."
            )

        # build runtime parameters
        runtime_parameters = runtime_parameters or {}
        if batch_data:
            runtime_parameters["batch_data"] = batch_data
        elif query:
            runtime_parameters["query"] = query
        elif path:
            runtime_parameters["path"] = path

        if batch_identifiers is None:
            batch_identifiers = kwargs
        else:
            # TODO: Raise a warning if kwargs exist
            pass

        RuntimeBatchRequest._validate_runtime_batch_request_specific_init_parameters(
            runtime_parameters=runtime_parameters,
            batch_identifiers=batch_identifiers,
            batch_spec_passthrough=batch_spec_passthrough,
        )

        batch_request_as_dict: BlockConfigBatchRequestTypedDict = {
            "datasource_name": datasource_name,
            "data_connector_name": data_connector_name,
            "data_asset_name": data_asset_name,
            "runtime_parameters": runtime_parameters,
            "batch_identifiers": batch_identifiers,
            "batch_spec_passthrough": batch_spec_passthrough,  # type: ignore[typeddict-item] # could be None
        }

        deep_filter_properties_iterable(
            properties=batch_request_as_dict,  # type: ignore[call-overload] # TypeDict is dict at runtime
            inplace=True,
        )

        return RuntimeBatchRequest(**batch_request_as_dict)  # type: ignore[misc] # `data_connector_query` is not a valid kw arg.
    else:
        return None


def get_batch_request_from_acceptable_arguments(  # noqa: PLR0913 # FIXME CoP
    datasource_name: str | None = None,
    data_connector_name: str | None = None,
    data_asset_name: str | None = None,
    *,
    batch_request: BatchRequestBase | None = None,
    batch_data: Any | None = None,
    data_connector_query: dict | None = None,
    batch_identifiers: dict | None = None,
    limit: int | None = None,
    index: int | list | tuple | slice | str | None = None,
    custom_filter_function: Callable | None = None,
    batch_spec_passthrough: dict | None = None,
    sampling_method: str | None = None,
    sampling_kwargs: dict | None = None,
    partitioner_method: str | None = None,
    partitioner_kwargs: dict | None = None,
    runtime_parameters: dict | None = None,
    query: str | None = None,
    path: str | None = None,
    batch_filter_parameters: dict | None = None,
    batch_parameters: dict | BatchParameters | None = None,
    **kwargs,
) -> BatchRequest | RuntimeBatchRequest | FluentBatchRequest:
    """Obtain formal BatchRequest typed object from allowed attributes (supplied as arguments).
    This method applies only to the new (V3) Datasource schema.

    Args:
        datasource_name
        data_connector_name
        data_asset_name

        batch_request
        batch_data
        query
        path
        runtime_parameters
        data_connector_query
        batch_identifiers
        batch_filter_parameters

        limit
        index
        custom_filter_function

        sampling_method
        sampling_kwargs

        partitioner_method
        partitioner_kwargs

        batch_spec_passthrough

        batch_parameters

        **kwargs

    Returns:
        (BatchRequest, RuntimeBatchRequest or FluentBatchRequest) The formal BatchRequest, RuntimeBatchRequest or FluentBatchRequest object
    """  # noqa: E501 # FIXME CoP

    # block-style batch-request args, includes arguments for both runtime and basic batch requests
    block_config_args = {
        "batch_data": batch_data,
        "query": query,
        "path": path,
        "runtime_parameters": runtime_parameters,
        "data_connector_query": data_connector_query,
        "batch_identifiers": batch_identifiers,
        "batch_filter_parameters": batch_filter_parameters,
        "limit": limit,
        "index": index,
        "custom_filter_function": custom_filter_function,
        "sampling_method": sampling_method,
        "sampling_kwargs": sampling_kwargs,
        "partitioner_method": partitioner_method,
        "partitioner_kwargs": partitioner_kwargs,
        "batch_spec_passthrough": batch_spec_passthrough,
    }

    def block_style_args() -> list[str]:
        """Returns a list of the block-config batch request arguments"""
        return [k for k, v in block_config_args.items() if v]

    # ensure that the first parameter is datasource_name, which should be a str. This check prevents users  # noqa: E501 # FIXME CoP
    # from passing in batch_request as an unnamed parameter.
    if datasource_name and not isinstance(datasource_name, str):
        raise TypeError(  # noqa: TRY003 # FIXME CoP
            f"the first parameter, datasource_name, must be a str, not {type(datasource_name)}"
        )

    # if batch_request is specified, ignore all other arguments and return the batch_request
    if batch_request:
        if not isinstance(
            batch_request,
            (BatchRequest, RuntimeBatchRequest, _get_fluent_batch_request_class()),
        ):
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                "batch_request must be a BatchRequest, RuntimeBatchRequest, or a "
                f"fluent BatchRequest object, not {type(batch_request)}"
            )

        return batch_request

    # try to get a runtime batch request
    result: RuntimeBatchRequest | FluentBatchRequest | None
    result = _get_runtime_batch_request(
        datasource_name=datasource_name,  # type: ignore[arg-type] # could be None
        data_connector_name=data_connector_name,  # type: ignore[arg-type] # could be None
        data_asset_name=data_asset_name,  # type: ignore[arg-type] # could be None
        runtime_parameters=runtime_parameters,
        batch_identifiers=batch_identifiers,
        batch_spec_passthrough=batch_spec_passthrough,
        batch_data=batch_data,
        query=query,
        path=path,
        **kwargs,
    )
    if result:
        return result

    # try to get a fluent batch request
    if (datasource_name and data_asset_name) and not data_connector_name:
        block_args = block_style_args()
        if block_args:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                f"Arguments: {', '.join(block_args)} are not supported for Fluent Batch Requests. Block-config Requests require a data connector name"  # noqa: E501 # FIXME CoP
            )

        result = _get_fluent_batch_request_class()(
            datasource_name=datasource_name,
            data_asset_name=data_asset_name,
            options=batch_parameters,
        )
        return result

    # try to get a block batch request
    return _get_block_batch_request(
        datasource_name=datasource_name,
        data_connector_name=data_connector_name,
        data_asset_name=data_asset_name,
        batch_identifiers=batch_identifiers,
        batch_spec_passthrough=batch_spec_passthrough,
        batch_filter_parameters=batch_filter_parameters,
        data_connector_query=data_connector_query,
        limit=limit,
        index=index,
        custom_filter_function=custom_filter_function,
        sampling_method=sampling_method,
        sampling_kwargs=sampling_kwargs,
        partitioner_method=partitioner_method,
        partitioner_kwargs=partitioner_kwargs,
        **kwargs,
    )


def standardize_batch_request_display_ordering(
    batch_request: BlockConfigBatchRequestTypedDict,
) -> BlockConfigBatchRequestTypedDict:
    batch_request_as_dict: dict = safe_deep_copy(data=batch_request)
    datasource_name: str = batch_request_as_dict["datasource_name"]
    data_connector_name: str = batch_request_as_dict["data_connector_name"]
    data_asset_name: str = batch_request_as_dict["data_asset_name"]
    runtime_parameters: dict | None = batch_request_as_dict.get("runtime_parameters")
    batch_identifiers: dict | None = batch_request_as_dict.get("batch_identifiers")
    batch_request_as_dict.pop("datasource_name")
    batch_request_as_dict.pop("data_connector_name")
    batch_request_as_dict.pop("data_asset_name")
    # NOTE: AJB 20211217 The below conditionals should be refactored
    if runtime_parameters is not None:
        batch_request_as_dict.pop("runtime_parameters")
    if batch_identifiers is not None:
        batch_request_as_dict.pop("batch_identifiers")
    if runtime_parameters is not None and batch_identifiers is not None:
        batch_request_as_dict = {
            "datasource_name": datasource_name,
            "data_connector_name": data_connector_name,
            "data_asset_name": data_asset_name,
            "runtime_parameters": runtime_parameters,
            "batch_identifiers": batch_identifiers,
            **batch_request_as_dict,
        }
    elif runtime_parameters is not None and batch_identifiers is None:
        batch_request_as_dict = {
            "datasource_name": datasource_name,
            "data_connector_name": data_connector_name,
            "data_asset_name": data_asset_name,
            "runtime_parameters": runtime_parameters,
            **batch_request_as_dict,
        }
    elif runtime_parameters is None and batch_identifiers is not None:
        batch_request_as_dict = {
            "datasource_name": datasource_name,
            "data_connector_name": data_connector_name,
            "data_asset_name": data_asset_name,
            "batch_identifiers": batch_identifiers,
            **batch_request_as_dict,
        }
    else:
        batch_request_as_dict = {
            "datasource_name": datasource_name,
            "data_connector_name": data_connector_name,
            "data_asset_name": data_asset_name,
            **batch_request_as_dict,
        }

    return batch_request_as_dict  # type: ignore[return-value] # TODO: create a new object to return instead of popping


if TYPE_CHECKING:
    AnyBatch: TypeAlias = Union[Batch, FluentBatch]

if pyspark.DataFrame:  # type: ignore[truthy-function] # False if NotImported
    BatchDataUnion: TypeAlias = Union[BatchData, pd.DataFrame, pyspark.DataFrame]

    BatchDataType: TypeAlias = Union[Type[BatchData], Type[pd.DataFrame], Type[pyspark.DataFrame]]
else:
    BatchDataType = Union[Type[BatchData], Type[pd.DataFrame]]  # type: ignore[misc] # Cannot assign multiple types
    BatchDataUnion = Union[BatchData, pd.DataFrame]  # type: ignore[misc] # Cannot assign multiple types



from __future__ import annotations

import json
import logging
from typing import Type

import pytest

from great_expectations.checkpoint.actions import (
    APINotificationAction,
    EmailAction,
    MicrosoftTeamsNotificationAction,
    OpsgenieAlertAction,
    SlackNotificationAction,
    SNSNotificationAction,
    UpdateDataDocsAction,
    ValidationAction,
)

logger = logging.getLogger(__name__)


class TestActionSerialization:
    EXAMPLE_SLACK_WEBHOOK = "https://hooks.slack.com/services/test/slack/webhook"
    EXAMPLE_TEAMS_WEBHOOK = "https://hooks.microsoft.com/services/test/teams/webhook"
    EXAMPLE_API_KEY = "testapikey"
    EXAMPLE_SMTP_ADDRESS = "smtp.test.com"
    EXAMPLE_SMTP_PORT = 587
    EXAMPLE_EMAILS = "bob@gmail.com, jim@hotmail.com"
    EXAMPLE_SITE_NAMES = ["one_site", "two_site", "red_site", "blue_site"]
    EXAMPLE_SNS_TOPIC_ARN = "my_test_arn"
    EXAMPLE_URL = "http://www.example.com"

    ACTION_INIT_PARAMS = {
        SlackNotificationAction: {
            "name": "my_slack_action",
            "slack_webhook": EXAMPLE_SLACK_WEBHOOK,
        },
        MicrosoftTeamsNotificationAction: {
            "name": "my_teams_action",
            "teams_webhook": EXAMPLE_TEAMS_WEBHOOK,
        },
        OpsgenieAlertAction: {"name": "my_opsgenie_action", "api_key": EXAMPLE_API_KEY},
        EmailAction: {
            "name": "my_email_action",
            "smtp_address": EXAMPLE_SMTP_ADDRESS,
            "smtp_port": EXAMPLE_SMTP_PORT,
            "receiver_emails": EXAMPLE_EMAILS,
        },
        UpdateDataDocsAction: {"name": "my_data_docs_action", "site_names": EXAMPLE_SITE_NAMES},
        SNSNotificationAction: {"name": "my_sns_action", "sns_topic_arn": EXAMPLE_SNS_TOPIC_ARN},
        APINotificationAction: {"name": "my_api_action", "url": EXAMPLE_URL},
    }

    SERIALIZED_ACTIONS = {
        SlackNotificationAction: {
            "name": "my_slack_action",
            "notify_on": "all",
            "notify_with": None,
            "renderer": {
                "class_name": "SlackRenderer",
                "module_name": "great_expectations.render.renderer.slack_renderer",
            },
            "show_failed_expectations": False,
            "slack_channel": None,
            "slack_token": None,
            "slack_webhook": EXAMPLE_SLACK_WEBHOOK,
            "type": "slack",
        },
        MicrosoftTeamsNotificationAction: {
            "name": "my_teams_action",
            "notify_on": "all",
            "renderer": {
                "class_name": "MicrosoftTeamsRenderer",
                "module_name": "great_expectations.render.renderer.microsoft_teams_renderer",
            },
            "teams_webhook": EXAMPLE_TEAMS_WEBHOOK,
            "type": "microsoft",
        },
        OpsgenieAlertAction: {
            "name": "my_opsgenie_action",
            "api_key": EXAMPLE_API_KEY,
            "notify_on": "failure",
            "priority": "P3",
            "region": None,
            "renderer": {
                "class_name": "OpsgenieRenderer",
                "module_name": "great_expectations.render.renderer.opsgenie_renderer",
            },
            "tags": None,
            "type": "opsgenie",
        },
        EmailAction: {
            "name": "my_email_action",
            "notify_on": "all",
            "notify_with": None,
            "receiver_emails": EXAMPLE_EMAILS,
            "renderer": {
                "class_name": "EmailRenderer",
                "module_name": "great_expectations.render.renderer.email_renderer",
            },
            "sender_alias": None,
            "sender_login": None,
            "sender_password": None,
            "smtp_address": EXAMPLE_SMTP_ADDRESS,
            "smtp_port": str(EXAMPLE_SMTP_PORT),
            "type": "email",
            "use_ssl": None,
            "use_tls": None,
        },
        UpdateDataDocsAction: {
            "name": "my_data_docs_action",
            "site_names": EXAMPLE_SITE_NAMES,
            "type": "update_data_docs",
        },
        SNSNotificationAction: {
            "name": "my_sns_action",
            "sns_message_subject": None,
            "sns_topic_arn": EXAMPLE_SNS_TOPIC_ARN,
            "type": "sns",
        },
        APINotificationAction: {
            "name": "my_api_action",
            "type": "api",
            "url": EXAMPLE_URL,
        },
    }

    @pytest.mark.parametrize(
        "action_class, init_params",
        [(k, v) for k, v in ACTION_INIT_PARAMS.items()],
        ids=[k.__name__ for k in ACTION_INIT_PARAMS],
    )
    @pytest.mark.unit
    def test_action_serialization_and_deserialization(
        self,
        mock_context,
        action_class: Type[ValidationAction],
        init_params: dict,
    ):
        expected = self.SERIALIZED_ACTIONS[action_class]

        action = action_class(**init_params)
        json_dict = action.json()
        actual = json.loads(json_dict)

        assert actual == expected

    @pytest.mark.parametrize(
        "action_class, serialized_action",
        [(k, v) for k, v in SERIALIZED_ACTIONS.items()],
        ids=[k.__name__ for k in SERIALIZED_ACTIONS],
    )
    @pytest.mark.unit
    def test_action_deserialization(
        self, action_class: Type[ValidationAction], serialized_action: dict
    ):
        actual = action_class.parse_obj(serialized_action)
        assert isinstance(actual, action_class)



import pytest
import responses

from great_expectations.data_context._version_checker import _VersionChecker

# Set to some arbitrary value so tests will continue to work regardless of GX's actual version
_MOCK_PYPI_VERSION = "0.16.8"


@pytest.fixture
def enable_pypi_version_check():
    stashed_version = _VersionChecker._LATEST_GX_VERSION_CACHE
    _VersionChecker._LATEST_GX_VERSION_CACHE = None
    yield
    _VersionChecker._LATEST_GX_VERSION_CACHE = stashed_version


@pytest.mark.parametrize(
    "version,expected,status",
    [
        pytest.param("0.15.0", False, 200, id="outdated"),
        pytest.param(_MOCK_PYPI_VERSION, True, 200, id="up-to-date"),
        # packaging.version should take care of dirty hashes but worth checking against
        pytest.param(
            f"{_MOCK_PYPI_VERSION}+59.g1ff4de04d.dirty",
            True,
            200,
            id="up-to-date local dev",
        ),
        # If we error, we shouldn't raise a warning to the user
        pytest.param(_MOCK_PYPI_VERSION, True, 400, id="bad request"),
    ],
)
@pytest.mark.unit
@responses.activate
def test_check_if_using_latest_gx(
    enable_pypi_version_check, version: str, expected: bool, status: int, caplog
):
    pypi_payload = {"info": {"version": _MOCK_PYPI_VERSION}}
    responses.add(
        responses.GET,
        _VersionChecker._PYPI_GX_ENDPOINT,
        json=pypi_payload,
        status=status,
    )

    checker = _VersionChecker(version)
    actual = checker.check_if_using_latest_gx()

    assert actual is expected
    if not actual:
        assert "upgrade" in caplog.text



import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.compatibility import pydantic
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    ALL_DATA_SOURCES,
    JUST_PANDAS_DATA_SOURCES,
)

COL_A = "col_a"
COL_B = "col_b"


DATA = pd.DataFrame({COL_A: [1, 2, None], COL_B: ["a", "b", None]})
EMPTY_DATA = pd.DataFrame({COL_A: [], COL_B: []})


@parameterize_batch_for_data_sources(data_source_configs=ALL_DATA_SOURCES, data=DATA)
def test_golden_path(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectTableRowCountToBeBetween(min_value=2, max_value=4)
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectTableRowCountToBeBetween(min_value=None, max_value=None),
            id="vacuously_true",
        ),
        pytest.param(
            gxe.ExpectTableRowCountToBeBetween(min_value=2, max_value=None),
            id="just_min",
        ),
        pytest.param(
            gxe.ExpectTableRowCountToBeBetween(min_value=None, max_value=4),
            id="just_max",
        ),
        pytest.param(
            gxe.ExpectTableRowCountToBeBetween(min_value=3, max_value=5),
            id="inclusivity",
        ),
        pytest.param(
            gxe.ExpectTableRowCountToBeBetween(
                min_value=None, max_value=None, strict_min=True, strict_max=True
            ),
            id="strict_min_max_vacuously_true",
        ),
        pytest.param(
            gxe.ExpectTableRowCountToBeBetween(
                min_value=2, max_value=None, strict_min=True, strict_max=True
            ),
            id="strict_min_max_just_min",
        ),
        pytest.param(
            gxe.ExpectTableRowCountToBeBetween(
                min_value=None, max_value=4, strict_min=True, strict_max=True
            ),
            id="strict_min_max_just_max",
        ),
        pytest.param(
            gxe.ExpectTableRowCountToBeBetween(
                min_value=2, max_value=4, strict_min=True, strict_max=True
            ),
            id="strict_min_max_inclusive",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success(
    batch_for_datasource: Batch, expectation: gxe.ExpectTableRowCountToBeBetween
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=EMPTY_DATA)
def test_empty_data(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectTableRowCountToBeBetween(min_value=0, max_value=0)
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectTableRowCountToBeBetween(min_value=8, max_value=None),
            id="just_min",
        ),
        pytest.param(
            gxe.ExpectTableRowCountToBeBetween(min_value=None, max_value=1),
            id="just_max",
        ),
        pytest.param(
            gxe.ExpectTableRowCountToBeBetween(min_value=4, max_value=4),
            id="bad_range",
        ),
        pytest.param(
            gxe.ExpectTableRowCountToBeBetween(min_value=3, max_value=4, strict_min=True),
            id="strict_min_max_observed_same_as_min",
        ),
        pytest.param(
            gxe.ExpectTableRowCountToBeBetween(min_value=2, max_value=3, strict_max=True),
            id="strict_min_max_observed_same_as_max",
        ),
        pytest.param(
            gxe.ExpectTableRowCountToBeBetween(
                min_value=3, max_value=3, strict_min=True, strict_max=True
            ),
            id="strict_min_max_observed_same_as_min_and_max",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_failure(
    batch_for_datasource: Batch, expectation: gxe.ExpectTableRowCountToBeBetween
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert not result.success


@pytest.mark.unit
def test_valid_range() -> None:
    with pytest.raises(pydantic.ValidationError):
        gxe.ExpectTableRowCountToBeBetween(min_value=5, max_value=4)


@pytest.mark.unit
def test_valid_runtime_parameters() -> None:
    gxe.ExpectTableRowCountToBeBetween(
        min_value={"$PARAMETER": "param_min_value"},
        max_value={"$PARAMETER": "param_max_values"},
    )


@pytest.mark.unit
def test_invalid_runtime_parameters() -> None:
    with pytest.raises(pydantic.ValidationError):
        gxe.ExpectTableRowCountToBeBetween(
            min_value={"min_value": "param_min_value"},
            max_value={"max_value": "param_max_values"},
        )


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(True, True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_strict_min_(
    batch_for_datasource: Batch, suite_param_value: bool, expected_result: bool
) -> None:
    suite_param_key = "test_expect_table_row_count_to_be_between"
    expectation = gxe.ExpectTableRowCountToBeBetween(
        min_value=2,
        max_value=4,
        strict_min={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result


@pytest.mark.parametrize(
    "suite_param_value,expected_result",
    [
        pytest.param(True, True, id="success"),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success_with_suite_param_strict_max_(
    batch_for_datasource: Batch, suite_param_value: bool, expected_result: bool
) -> None:
    suite_param_key = "test_expect_table_row_count_to_be_between"
    expectation = gxe.ExpectTableRowCountToBeBetween(
        min_value=2,
        max_value=4,
        strict_max={"$PARAMETER": suite_param_key},
        result_format=ResultFormat.SUMMARY,
    )
    result = batch_for_datasource.validate(
        expectation, expectation_parameters={suite_param_key: suite_param_value}
    )
    assert result.success == expected_result



from __future__ import annotations

from typing import Literal, Optional, Union

from great_expectations.compatibility.pydantic import Field
from great_expectations.compatibility.typing_extensions import override
from great_expectations.datasource.fluent.data_asset.path.directory_asset import (
    DirectoryDataAsset,
)
from great_expectations.datasource.fluent.data_asset.path.file_asset import FileDataAsset
from great_expectations.datasource.fluent.data_asset.path.spark.spark_generic import (
    _SparkGenericFilePathAssetMixin,
)


class ORCAssetBase(_SparkGenericFilePathAssetMixin):
    # The options below are available as of spark v3.4.0
    # See https://spark.apache.org/docs/latest/sql-data-sources-orc.html for more info.
    merge_schema: Optional[Union[bool, str]] = Field(False, alias="mergeSchema")

    class Config:
        extra = "forbid"

        allow_population_by_field_name = True

    @classmethod
    @override
    def _get_reader_method(cls) -> str:
        return "orc"

    @override
    def _get_reader_options_include(self) -> set[str]:
        """These options are available as of spark v3.4.0

        See https://spark.apache.org/docs/latest/sql-data-sources-orc.html for more info.
        """
        return super()._get_reader_options_include().union({"merge_schema"})


class ORCAsset(FileDataAsset, ORCAssetBase):
    type: Literal["orc"] = "orc"


class DirectoryORCAsset(DirectoryDataAsset, ORCAssetBase):
    type: Literal["directory_orc"] = "directory_orc"

    @classmethod
    @override
    def _get_reader_method(cls) -> str:
        return "orc"

    @override
    def _get_reader_options_include(self) -> set[str]:
        """These options are available as of spark v3.4.0

        See https://spark.apache.org/docs/latest/sql-data-sources-orc.html for more info.
        """
        return (
            super()._get_reader_options_include()
            | super(DirectoryDataAsset, self)._get_reader_options_include()
        )



from __future__ import annotations

import logging
import urllib.parse
import warnings
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Literal,
    Mapping,
    Optional,
    TypedDict,
)

from great_expectations.compatibility.pydantic import AnyUrl, SecretStr, parse_obj_as
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.config_substitutor import TEMPLATE_STR_REGEX

if TYPE_CHECKING:
    from typing_extensions import Self, TypeAlias

    from great_expectations.core.config_provider import _ConfigurationProvider
    from great_expectations.datasource.fluent import Datasource

LOGGER = logging.getLogger(__name__)


class ConfigStr(SecretStr):
    """
    Special type that enables great_expectation config variable substitution.

    To enable config substitution for Fluent Datasources or DataAsset fields must be of
    the `ConfigStr` type, or a union containing this type.

    Note: this type is meant to used as part of pydantic model.
    To use this outside of a model see the pydantic docs below.
    https://docs.pydantic.dev/usage/models/#parsing-data-into-a-specified-type
    """

    def __init__(
        self,
        template_str: str,
    ) -> None:
        self.template_str: str = template_str
        self._secret_value = template_str  # for compatibility with SecretStr

    def get_config_value(self, config_provider: _ConfigurationProvider) -> str:
        """
        Resolve the config template string to its string value according to the passed
        _ConfigurationProvider.
        """
        LOGGER.info(f"Substituting '{self}'")
        return config_provider.substitute_config(self.template_str)

    def _display(self) -> str:
        return str(self)

    @override
    def __str__(self) -> str:
        return self.template_str

    @override
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._display()!r})"

    @classmethod
    def str_contains_config_template(cls, v: str) -> bool:
        """
        Returns True if the input string contains a config template string.
        """
        return TEMPLATE_STR_REGEX.search(v) is not None

    @classmethod
    def validate_template_str_format(cls, v: str) -> str | None:
        if cls.str_contains_config_template(v):
            return v
        raise ValueError(
            cls.__name__
            + " - contains no config template strings in the format"
            + r" '${MY_CONFIG_VAR}' or '$MY_CONFIG_VAR'"
        )

    @classmethod
    def __get_validators__(cls):
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield cls.validate_template_str_format
        yield cls.validate

    @classmethod
    @override
    def __modify_schema__(cls, field_schema: dict) -> None:
        """Update the generated schema when used in a pydantic model."""
        SecretStr.__modify_schema__(field_schema)
        field_schema.update(
            {
                "description": "Contains config templates"
                " to be substituted at runtime. Runtime values will never be serialized.",
                "pattern": ".*" + TEMPLATE_STR_REGEX.pattern + ".*",
                "examples": [
                    "hello_${NAME}",
                    "${MY_CFG_VAR}",
                ],
            }
        )


UriParts: TypeAlias = Literal[  # https://docs.pydantic.dev/1.10/usage/types/#url-properties
    "scheme", "host", "user", "password", "port", "path", "query", "fragment", "tld"
]


class UriPartsDict(TypedDict, total=False):
    scheme: str
    user: str | None
    password: str | None
    ipv4: str | None
    ipv6: str | None
    domain: str | None
    port: str | None
    path: str | None
    query: str | None
    fragment: str | None


class ConfigUri(AnyUrl, ConfigStr):  # type: ignore[misc] # Mixin "validate" signature mismatch
    """
    Special type that enables great_expectation config variable substitution for the
    `user` and `password` section of a URI.

    Example:
    ```
    "snowflake://${MY_USER}:${MY_PASSWORD}@account/database/schema/table"
    ```

    Note: this type is meant to used as part of pydantic model.
    To use this outside of a model see the pydantic docs below.
    https://docs.pydantic.dev/usage/models/#parsing-data-into-a-specified-type
    """

    ALLOWED_SUBSTITUTIONS: ClassVar[set[UriParts]] = {"user", "password"}

    min_length: int = 1
    max_length: int = 2**16

    def __init__(  # noqa: PLR0913 # for compatibility with AnyUrl
        self,
        template_str: str,
        *,
        scheme: str,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        tld: Optional[str] = None,
        host_type: str = "domain",
        port: Optional[str] = None,
        path: Optional[str] = None,
        query: Optional[str] = None,
        fragment: Optional[str] = None,
    ) -> None:
        if template_str:  # may have already been set in __new__
            self.template_str: str = template_str
        self._secret_value = template_str  # for compatibility with SecretStr
        super().__init__(
            template_str,
            scheme=scheme,
            user=user,
            password=password,
            host=host,
            tld=tld,
            host_type=host_type,
            port=port,
            path=path,
            query=query,
            fragment=fragment,
        )

    def __new__(cls: type[Self], template_str: Optional[str], **kwargs) -> Self:
        """custom __new__ for compatibility with pydantic.parse_obj_as()"""
        built_url = cls.build(**kwargs) if template_str is None else template_str
        instance = str.__new__(cls, built_url)
        instance.template_str = str(instance)
        return instance

    @property
    def params(self) -> dict[str, list[str]]:
        """The query parameters as a dictionary."""
        if not self.query:
            return {}
        return urllib.parse.parse_qs(self.query)

    @classmethod
    @override
    def validate_parts(cls, parts: UriPartsDict, validate_port: bool = True) -> UriPartsDict:
        """
        Ensure that only the `user` and `password` parts have config template strings.
        Also validate that all parts of the URI are valid.
        """
        allowed_substitutions = sorted(cls.ALLOWED_SUBSTITUTIONS)

        for name, part in parts.items():
            if not part:
                continue
            if (
                cls.str_contains_config_template(part)  # type: ignore[arg-type] # is str
                and name not in cls.ALLOWED_SUBSTITUTIONS
            ):
                raise ValueError(  # noqa: TRY003 # FIXME CoP
                    f"Only {', '.join(allowed_substitutions)} may use config substitution; '{name}'"
                    " substitution not allowed"
                )
        return AnyUrl.validate_parts(parts, validate_port)

    @override
    def get_config_value(self, config_provider: _ConfigurationProvider) -> AnyUrl:
        """
        Resolve the config template string to its string value according to the passed
        _ConfigurationProvider.
        Parse the resolved URI string into an `AnyUrl` object.
        """
        LOGGER.info(f"Substituting '{self}'")
        raw_value = config_provider.substitute_config(self.template_str)
        return parse_obj_as(AnyUrl, raw_value)

    @classmethod
    def __get_validators__(cls):
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield ConfigStr.validate_template_str_format
        yield cls.validate  # equivalent to AnyUrl.validate

    @classmethod
    @override
    def __modify_schema__(cls, field_schema: dict) -> None:
        """Update the generated schema when used in a pydantic model."""
        ConfigStr.__modify_schema__(field_schema)
        AnyUrl.__modify_schema__(field_schema)
        field_schema.update(
            {
                "description": "Contains config templates for user:password in a URI"
                " to be substituted at runtime. Runtime values will never be serialized.",
                "examples": [
                    "snowflake://dickens:${PASSWORD}@host/db/schema",
                    "snowflake://${USER}:${PASSWORD}@host/db/schema",
                ],
            }
        )


def _check_config_substitutions_needed(
    datasource: Datasource,
    options: Mapping,
    raise_warning_if_provider_not_present: bool,
) -> set[str]:
    """
    Given a Datasource and a dict-like mapping type return the keys whose value is a `ConfigStr` type.
    Optionally raise a warning if config substitution is needed but impossible due to a missing `_config_provider`.
    """  # noqa: E501 # FIXME CoP
    need_config_subs: set[str] = {k for (k, v) in options.items() if isinstance(v, ConfigStr)}
    if (
        need_config_subs
        and raise_warning_if_provider_not_present
        and not datasource._config_provider
    ):
        warnings.warn(
            f"config variables '{','.join(need_config_subs)}' need substitution but no `_ConfigurationProvider` is present"  # noqa: E501 # FIXME CoP
        )
    return need_config_subs



from __future__ import annotations

from typing import Literal, Optional

from great_expectations.compatibility.pydantic import Field
from great_expectations.compatibility.typing_extensions import override
from great_expectations.datasource.fluent.data_asset.path.directory_asset import (
    DirectoryDataAsset,
)
from great_expectations.datasource.fluent.data_asset.path.file_asset import FileDataAsset
from great_expectations.datasource.fluent.data_asset.path.path_data_asset import (
    PathDataAsset,
)


class DeltaAssetBase(PathDataAsset):
    # The options below are available as of 2023-05-12
    # See https://docs.databricks.com/delta/tutorial.html for more info.

    timestamp_as_of: Optional[str] = Field(None, alias="timestampAsOf")
    version_as_of: Optional[str] = Field(None, alias="versionAsOf")

    class Config:
        extra = "forbid"

        allow_population_by_field_name = True

    @classmethod
    @override
    def _get_reader_method(cls) -> str:
        return "delta"

    @override
    def _get_reader_options_include(self) -> set[str]:
        """The options below are available as of 2023-05-12

        See https://docs.databricks.com/delta/tutorial.html for more info.
        """
        return {"timestamp_as_of", "version_as_of"}


class DeltaAsset(FileDataAsset, DeltaAssetBase):
    type: Literal["delta"] = "delta"


class DirectoryDeltaAsset(DirectoryDataAsset, DeltaAssetBase):
    type: Literal["directory_delta"] = "directory_delta"

    @classmethod
    @override
    def _get_reader_method(cls) -> str:
        return "delta"

    @override
    def _get_reader_options_include(self) -> set[str]:
        """The options below are available as of 2023-05-12

        See https://docs.databricks.com/delta/tutorial.html for more info.
        """
        return (
            super()._get_reader_options_include()
            | super(DirectoryDataAsset, self)._get_reader_options_include()
        )



from __future__ import annotations

from itertools import zip_longest
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Optional, Type, Union

from great_expectations.compatibility import pydantic
from great_expectations.core.suite_parameters import (
    SuiteParameterDict,  # noqa: TC001 # FIXME CoP
)
from great_expectations.expectations.expectation import (
    BatchExpectation,
    render_suite_parameter_string,
)
from great_expectations.expectations.metadata_types import DataQualityIssues, SupportedDataSources
from great_expectations.render import (
    AtomicDiagnosticRendererType,
    LegacyRendererType,
    RenderedAtomicContent,
    RenderedStringTemplateContent,
    renderedAtomicValueSchema,
)
from great_expectations.render.renderer.observed_value_renderer import (
    _prepare_params_for_list_comparison,
)
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererValueType,
)
from great_expectations.render.util import substitute_none_for_missing

if TYPE_CHECKING:
    from great_expectations.core import (
        ExpectationValidationResult,
    )
    from great_expectations.execution_engine import ExecutionEngine
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )


EXPECTATION_SHORT_DESCRIPTION = "Expect the columns in a table to exactly match a specified list."
COLUMN_LIST_DESCRIPTION = "The column names, in the correct order."
SUPPORTED_DATA_SOURCES = [
    SupportedDataSources.PANDAS.value,
    SupportedDataSources.SPARK.value,
    SupportedDataSources.SQLITE.value,
    SupportedDataSources.POSTGRESQL.value,
    SupportedDataSources.MYSQL.value,
    SupportedDataSources.MSSQL.value,
    SupportedDataSources.BIGQUERY.value,
    SupportedDataSources.SNOWFLAKE.value,
    SupportedDataSources.DATABRICKS.value,
    SupportedDataSources.REDSHIFT.value,
]
DATA_QUALITY_ISSUES = [DataQualityIssues.SCHEMA.value]


class ExpectTableColumnsToMatchOrderedList(BatchExpectation):
    __doc__ = f"""{EXPECTATION_SHORT_DESCRIPTION}

    ExpectTableColumnsToMatchOrderedList is a \
    Batch Expectation.

    BatchExpectations are one of the most common types of Expectation.
    They are evaluated for an entire Batch, and answer a semantic question about the Batch itself.

    Args:
        column_list (list of str): \
            {COLUMN_LIST_DESCRIPTION}

    Other Parameters:
        result_format (str or None): \
            Which output mode to use: BOOLEAN_ONLY, BASIC, COMPLETE, or SUMMARY. \
            For more detail, see [result_format](https://docs.greatexpectations.io/docs/reference/expectations/result_format).
        catch_exceptions (boolean or None): \
            If True, then catch exceptions and include them as part of the result object. \
            For more detail, see [catch_exceptions](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#catch_exceptions).
        meta (dict or None): \
            A JSON-serializable dictionary (nesting allowed) that will be included in the output without \
            modification. For more detail, see [meta](https://docs.greatexpectations.io/docs/reference/expectations/standard_arguments/#meta).

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

        Exact fields vary depending on the values passed to result_format, catch_exceptions, and meta.

    Supported Data Sources:
        [{SUPPORTED_DATA_SOURCES[0]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[1]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[2]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[3]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[4]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[5]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[6]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[7]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[8]}](https://docs.greatexpectations.io/docs/application_integration_support/)

    Data Quality Issues:
        {DATA_QUALITY_ISSUES[0]}

    Example Data:
                test 	test2
            0 	1.00 	2
            1 	2.30 	5
            2 	4.33 	0

    Code Examples:
        Passing Case:
            Input:
                ExpectTableColumnsToMatchOrderedList(
                    column_list=["test", "test2"]
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "observed_value": [
                      "test",
                      "test2"
                    ]
                  }},
                  "meta": {{}},
                  "success": true
                }}

        Failing Case:
            Input:
                ExpectTableColumnsToMatchOrderedList(
                    column_list=["test2", "test", "test3"]
            )

            Output:
                {{
                  "exception_info": {{
                    "raised_exception": false,
                    "exception_traceback": null,
                    "exception_message": null
                  }},
                  "result": {{
                    "observed_value": [
                      "Unnamed: 0",
                      "test",
                      "test2"
                    ],
                    "details": {{
                      "mismatched": [
                        {{
                          "Expected Column Position": 1,
                          "Expected": "test2",
                          "Found": "test"
                        }},
                        {{
                          "Expected Column Position": 2,
                          "Expected": "test",
                          "Found": "test2"
                        }},
                        {{
                          "Expected Column Position": 3,
                          "Expected": "test3",
                          "Found": null
                        }}
                      ]
                    }}
                  }},
                  "meta": {{}},
                  "success": false
                }}
    """  # noqa: E501 # FIXME CoP

    column_list: Union[list, set, SuiteParameterDict, None] = pydantic.Field(
        description=COLUMN_LIST_DESCRIPTION
    )

    library_metadata: ClassVar[Dict[str, Union[str, list, bool]]] = {
        "maturity": "production",
        "tags": ["core expectation", "table expectation"],
        "contributors": ["@great_expectations"],
        "requirements": [],
        "has_full_test_suite": True,
        "manually_reviewed_code": True,
    }
    _library_metadata = library_metadata

    metric_dependencies = ("table.columns",)
    success_keys = ("column_list",)
    domain_keys = (
        "batch_id",
        "row_condition",
        "condition_parser",
    )
    args_keys = ("column_list",)

    class Config:
        title = "Expect table columns to match ordered list"

        @staticmethod
        def schema_extra(
            schema: Dict[str, Any], model: Type[ExpectTableColumnsToMatchOrderedList]
        ) -> None:
            BatchExpectation.Config.schema_extra(schema, model)
            schema["properties"]["metadata"]["properties"].update(
                {
                    "data_quality_issues": {
                        "title": "Data Quality Issues",
                        "type": "array",
                        "const": DATA_QUALITY_ISSUES,
                    },
                    "library_metadata": {
                        "title": "Library Metadata",
                        "type": "object",
                        "const": model._library_metadata,
                    },
                    "short_description": {
                        "title": "Short Description",
                        "type": "string",
                        "const": EXPECTATION_SHORT_DESCRIPTION,
                    },
                    "supported_data_sources": {
                        "title": "Supported Data Sources",
                        "type": "array",
                        "const": SUPPORTED_DATA_SOURCES,
                    },
                }
            )

    @classmethod
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        renderer_configuration.add_param(name="column_list", param_type=RendererValueType.ARRAY)
        params = renderer_configuration.params

        if not params.column_list:
            template_str = (
                "Must have a list of columns in a specific order, but that order is not specified."
            )
        else:
            template_str = "Must have these columns in this order: "
            array_param_name = "column_list"
            param_prefix = "column_list_"
            renderer_configuration = cls._add_array_params(
                array_param_name=array_param_name,
                param_prefix=param_prefix,
                renderer_configuration=renderer_configuration,
            )
            template_str += cls._get_array_string(
                array_param_name=array_param_name,
                param_prefix=param_prefix,
                renderer_configuration=renderer_configuration,
            )

        renderer_configuration.template_str = template_str

        return renderer_configuration

    @classmethod
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    @render_suite_parameter_string
    def _prescriptive_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        runtime_configuration = runtime_configuration or {}
        _ = runtime_configuration.get("include_column_name") is not False
        styling = runtime_configuration.get("styling")
        params = substitute_none_for_missing(configuration.kwargs, ["column_list"])

        if params["column_list"] is None:
            template_str = (
                "Must have a list of columns in a specific order, but that order is not specified."
            )

        else:
            template_str = "Must have these columns in this order: "
            for idx in range(len(params["column_list"]) - 1):
                template_str += f"$column_list_{idx!s}, "
                params[f"column_list_{idx!s}"] = params["column_list"][idx]

            last_idx = len(params["column_list"]) - 1
            template_str += f"$column_list_{last_idx!s}"
            params[f"column_list_{last_idx!s}"] = params["column_list"][last_idx]

        return [
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": template_str,
                        "params": params,
                        "styling": styling,
                    },
                }
            )
        ]

    def _validate(
        self,
        metrics: Dict,
        runtime_configuration: Optional[dict] = None,
        execution_engine: Optional[ExecutionEngine] = None,
    ):
        # Obtaining columns and ordered list for sake of comparison
        expected_column_list = self._get_success_kwargs().get("column_list")
        actual_column_list = metrics.get("table.columns")

        if expected_column_list is None or list(actual_column_list) == list(expected_column_list):
            return {
                "success": True,
                "result": {"observed_value": list(actual_column_list)},
            }
        else:
            # In the case of differing column lengths between the defined expectation and the observed column set, the  # noqa: E501 # FIXME CoP
            # max is determined to generate the column_index.
            number_of_columns = max(len(expected_column_list), len(actual_column_list))
            column_index = range(number_of_columns)

            # Create a list of the mismatched details
            compared_lists = list(
                zip_longest(column_index, list(expected_column_list), list(actual_column_list))
            )
            mismatched = [
                {"Expected Column Position": i, "Expected": k, "Found": v}
                for i, k, v in compared_lists
                if k != v
            ]
            return {
                "success": False,
                "result": {
                    "observed_value": list(actual_column_list),
                    "details": {"mismatched": mismatched},
                },
            }

    @classmethod
    @renderer(renderer_type=AtomicDiagnosticRendererType.OBSERVED_VALUE)
    def _atomic_diagnostic_observed_value(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
    ) -> RenderedAtomicContent:
        renderer_configuration = RendererConfiguration(
            configuration=configuration,
            result=result,
            runtime_configuration=runtime_configuration,
        )
        expected_param_prefix = "exp__"
        expected_param_name = "expected_value"
        ov_param_prefix = "ov__"
        ov_param_name = "observed_value"

        renderer_configuration.add_param(
            name=expected_param_name,
            param_type=RendererValueType.ARRAY,
            value=renderer_configuration.result.expectation_config.kwargs.get("column_list", []),
        )
        renderer_configuration = cls._add_array_params(
            array_param_name=expected_param_name,
            param_prefix=expected_param_prefix,
            renderer_configuration=renderer_configuration,
        )

        renderer_configuration.add_param(
            name=ov_param_name,
            param_type=RendererValueType.ARRAY,
            value=result.result.get("observed_value"),
        )
        renderer_configuration = cls._add_array_params(
            array_param_name=ov_param_name,
            param_prefix=ov_param_prefix,
            renderer_configuration=renderer_configuration,
        )

        renderer_configuration.template_str = _prepare_params_for_list_comparison(
            params=renderer_configuration.params,
            expected_prefix=expected_param_prefix,
            observed_prefix=ov_param_prefix,
        )

        value_obj = renderedAtomicValueSchema.load(
            {
                "template": renderer_configuration.template_str,
                "params": renderer_configuration.params.dict(),
                "meta_notes": renderer_configuration.meta_notes,
                "schema": {"type": "com.superconductive.rendered.string"},
            }
        )
        return RenderedAtomicContent(
            name=AtomicDiagnosticRendererType.OBSERVED_VALUE,
            value=value_obj,
            value_type="StringValueType",
        )



from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING, Callable, Final

import pact
import pytest

from tests.integration.cloud.rest_contracts.conftest import (
    EXISTING_ORGANIZATION_ID,
    ContractInteraction,
)

if TYPE_CHECKING:
    from tests.integration.cloud.rest_contracts.conftest import PactBody


NON_EXISTENT_DATASOURCE_ID: Final[str] = "6ed9a340-8469-4ee2-a300-ffbe5d09b49d"

EXISTING_DATASOURCE_ID: Final[str] = "15da041b-328e-44f7-892e-2bfd1a887ef8"


POST_DATASOURCE_MIN_RESPONSE_BODY: Final[PactBody] = {
    "data": pact.Like(
        {
            "id": pact.Format().uuid,
            "attributes": {
                "datasource_config": {},
            },
        },
    )
}

GET_DATASOURCE_MIN_RESPONSE_BODY: Final[PactBody] = {
    "data": pact.Like(
        {
            "id": pact.Format().uuid,
            "type": "pandas",
            "attributes": {
                "datasource_config": {
                    "assets": [],
                },
            },
        },
    )
}


@pytest.mark.cloud
@pytest.mark.parametrize(
    "contract_interaction",
    [
        # ContractInteraction(
        #     method="POST",
        #     request_path=pathlib.Path(
        #         "/",
        #         "organizations",
        #         EXISTING_ORGANIZATION_ID,
        #         "datasources",
        #     ),
        #     upon_receiving="a request to add a Data Source",
        #     given="the Data Source does not exist",
        #     response_status=200,
        #     response_body=POST_DATASOURCE_MIN_RESPONSE_BODY,
        # ),
        ContractInteraction(
            method="GET",
            request_path=pathlib.Path(
                "/",
                "organizations",
                EXISTING_ORGANIZATION_ID,
                "datasources",
                EXISTING_DATASOURCE_ID,
            ),
            upon_receiving="a request to get a Data Source",
            given="the Data Source exists",
            response_status=200,
            response_body=GET_DATASOURCE_MIN_RESPONSE_BODY,
        ),
    ],
)
def test_datasource(
    contract_interaction: ContractInteraction,
    run_rest_api_pact_test: Callable[[ContractInteraction], None],
) -> None:
    run_rest_api_pact_test(contract_interaction)



from __future__ import annotations

import copy
import json
import logging
from copy import deepcopy
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    List,
    Mapping,
    Optional,
    Type,
    Union,
)

from marshmallow import Schema, ValidationError, fields, post_dump, post_load, pre_dump
from typing_extensions import TypedDict

from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.core.suite_parameters import (
    build_suite_parameters,
)
from great_expectations.exceptions import (
    ExpectationNotFoundError,
    InvalidExpectationConfigurationError,
    InvalidExpectationKwargsError,
)
from great_expectations.expectations.registry import get_expectation_impl
from great_expectations.render import RenderedAtomicContent, RenderedAtomicContentSchema
from great_expectations.types import SerializableDictDot
from great_expectations.util import (
    convert_to_json_serializable,  # noqa: TID251 # FIXME CoP
    ensure_json_serializable,  # noqa: TID251 # FIXME CoP
)

if TYPE_CHECKING:
    from great_expectations.alias_types import JSONValues
    from great_expectations.data_context import AbstractDataContext
    from great_expectations.expectations.expectation import Expectation
logger = logging.getLogger(__name__)


def parse_result_format(result_format: Union[str, dict]) -> dict:
    """This is a simple helper utility that can be used to parse a string result_format into the dict format used
    internally by great_expectations. It is not necessary but allows shorthand for result_format in cases where
    there is no need to specify a custom partial_unexpected_count."""  # noqa: E501 # FIXME CoP
    if isinstance(result_format, str):
        result_format = {
            "result_format": result_format,
            "partial_unexpected_count": 20,
            "include_unexpected_rows": False,
        }
    else:
        if "include_unexpected_rows" in result_format and "result_format" not in result_format:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                "When using `include_unexpected_rows`, `result_format` must be explicitly specified"
            )

        if "partial_unexpected_count" not in result_format:
            result_format["partial_unexpected_count"] = 20

        if "include_unexpected_rows" not in result_format:
            result_format["include_unexpected_rows"] = False

    return result_format


class ExpectationContext(SerializableDictDot):
    def __init__(self, description: Optional[str] = None) -> None:
        self._description = description

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, value) -> None:
        self._description = value


class ExpectationContextSchema(Schema):
    description = fields.String(required=False, allow_none=True)

    @post_load
    def make_expectation_context(self, data, **kwargs):
        return ExpectationContext(**data)


class KWargDetailsDict(TypedDict):
    domain_kwargs: tuple[str, ...]
    success_kwargs: tuple[str, ...]
    default_kwarg_values: dict[str, str | bool | float | None]


# NOTE: This class is a legacy class that is being replaced by the new class-first expectations.
#       While it is still used in some places (namely around serialization), it is being phased out.
#       Please utilize Expectation objects directly when possible.
class ExpectationConfiguration(SerializableDictDot):
    """Defines the parameters and name of a specific Expectation.

    Args:
        type: The name of the expectation class to use in snake case, e.g. `expect_column_values_to_not_be_null`.
        kwargs: The keyword arguments to pass to the expectation class.
        meta: A dictionary of metadata to attach to the expectation.
        notes: Notes about this expectation.
        description: The description of the expectation. This will be rendered instead of the default template.
        success_on_last_run: Whether the expectation succeeded on the last run.
        id: The corresponding GX Cloud ID for the expectation.
        expectation_context: The context for the expectation.
        rendered_content: Rendered content for the expectation.
    Raises:
        InvalidExpectationConfigurationError: If `expectation_type` arg is not a str.
        InvalidExpectationConfigurationError: If `kwargs` arg is not a dict.
        InvalidExpectationKwargsError: If domain kwargs are missing.
        ValueError: If a `domain_type` cannot be determined.
    """  # noqa: E501 # FIXME CoP

    runtime_kwargs: ClassVar[tuple[str, ...]] = (
        "result_format",
        "catch_exceptions",
    )

    def __init__(  # noqa: PLR0913 # FIXME CoP
        self,
        type: str,
        kwargs: dict,
        meta: Optional[dict] = None,
        notes: str | list[str] | None = None,
        description: str | None = None,
        success_on_last_run: Optional[bool] = None,
        id: Optional[str] = None,
        expectation_context: Optional[ExpectationContext] = None,
        rendered_content: Optional[List[RenderedAtomicContent]] = None,
    ) -> None:
        if not isinstance(type, str):
            raise InvalidExpectationConfigurationError("expectation_type must be a string")  # noqa: TRY003 # FIXME CoP
        self._type = type
        if not isinstance(kwargs, dict):
            raise InvalidExpectationConfigurationError(  # noqa: TRY003 # FIXME CoP
                "expectation configuration kwargs must be a dict."
            )
        self._kwargs = kwargs
        # the kwargs before suite parameters are evaluated
        self._raw_kwargs: dict[str, Any] | None = None
        if meta is None:
            meta = {}
        # We require meta information to be serializable, but do not convert until necessary
        ensure_json_serializable(meta)
        self.meta = meta
        self.notes = notes
        self.description = description
        self.success_on_last_run = success_on_last_run
        self._id = id
        self._expectation_context = expectation_context
        self._rendered_content = rendered_content

    def process_suite_parameters(
        self,
        suite_parameters,
        interactive_evaluation: bool = True,
        data_context: Optional[AbstractDataContext] = None,
    ) -> None:
        if not self._raw_kwargs:
            suite_args, _ = build_suite_parameters(
                expectation_args=self._kwargs,
                suite_parameters=suite_parameters,
                interactive_evaluation=interactive_evaluation,
                data_context=data_context,
            )

            self._raw_kwargs = self._kwargs
            self._kwargs = suite_args
        else:
            logger.debug("suite_parameters have already been built on this expectation")

    def get_raw_configuration(self) -> ExpectationConfiguration:
        # return configuration without substituted suite parameters
        raw_config = deepcopy(self)
        if raw_config._raw_kwargs is not None:
            raw_config._kwargs = raw_config._raw_kwargs
            raw_config._raw_kwargs = None

        return raw_config

    @property
    def id(self) -> Optional[str]:
        return self._id

    @id.setter
    def id(self, value: str) -> None:
        self._id = value

    @property
    def expectation_context(self) -> Optional[ExpectationContext]:
        return self._expectation_context

    @expectation_context.setter
    def expectation_context(self, value: ExpectationContext) -> None:
        self._expectation_context = value

    @property
    def type(self) -> str:
        return self._type

    @property
    def kwargs(self) -> dict:
        return self._kwargs

    @kwargs.setter
    def kwargs(self, value: dict) -> None:
        self._kwargs = value

    @property
    def rendered_content(self) -> Optional[List[RenderedAtomicContent]]:
        return self._rendered_content

    @rendered_content.setter
    def rendered_content(self, value: Optional[List[RenderedAtomicContent]]) -> None:
        self._rendered_content = value

    def _get_default_custom_kwargs(self) -> KWargDetailsDict:
        # NOTE: this is a holdover until class-first expectations control their
        # defaults, and so defaults are inherited.
        if self.type.startswith("expect_column_pair"):
            return {
                "domain_kwargs": (
                    "column_A",
                    "column_B",
                    "row_condition",
                    "condition_parser",
                ),
                # NOTE: this is almost certainly incomplete; subclasses should override
                "success_kwargs": (),
                "default_kwarg_values": {
                    "column_A": None,
                    "column_B": None,
                    "row_condition": None,
                    "condition_parser": None,
                },
            }
        elif self.type.startswith("expect_column"):
            return {
                "domain_kwargs": ("column", "row_condition", "condition_parser"),
                # NOTE: this is almost certainly incomplete; subclasses should override
                "success_kwargs": (),
                "default_kwarg_values": {
                    "column": None,
                    "row_condition": None,
                    "condition_parser": None,
                },
            }

        logger.warning("Requested kwargs for an unrecognized expectation.")
        return {
            "domain_kwargs": (),
            # NOTE: this is almost certainly incomplete; subclasses should override
            "success_kwargs": (),
            "default_kwarg_values": {},
        }

    def get_domain_kwargs(self) -> dict:
        default_kwarg_values: dict[str, Any] = {}
        try:
            impl = self._get_expectation_impl()
        except ExpectationNotFoundError:
            expectation_kwargs_dict = self._get_default_custom_kwargs()
            domain_keys = expectation_kwargs_dict["domain_kwargs"]
        else:
            domain_keys = impl.domain_keys
            default_kwarg_values = self._get_expectation_class_defaults()

        domain_kwargs = {
            key: self.kwargs.get(key, default_kwarg_values.get(key)) for key in domain_keys
        }
        missing_kwargs = set(domain_keys) - set(domain_kwargs.keys())
        if missing_kwargs:
            raise InvalidExpectationKwargsError(f"Missing domain kwargs: {list(missing_kwargs)}")  # noqa: TRY003 # FIXME CoP

        return domain_kwargs

    def get_success_kwargs(self) -> dict:
        """Gets the success and domain kwargs for this ExpectationConfiguration.

        Raises:
            ExpectationNotFoundError: If the expectation implementation is not found.

        Returns:
            A dictionary with the success and domain kwargs of an expectation.
        """
        default_kwarg_values: Mapping[str, str | bool | float | object | None]
        try:
            impl = self._get_expectation_impl()
        except ExpectationNotFoundError:
            expectation_kwargs_dict = self._get_default_custom_kwargs()
            default_kwarg_values = expectation_kwargs_dict.get("default_kwarg_values", {})
            success_keys = expectation_kwargs_dict["success_kwargs"]
        else:
            success_keys = impl.success_keys
            default_kwarg_values = self._get_expectation_class_defaults()

        domain_kwargs = self.get_domain_kwargs()
        success_kwargs = {
            key: self.kwargs.get(key, default_kwarg_values.get(key)) for key in success_keys
        }
        success_kwargs.update(domain_kwargs)

        return success_kwargs

    def get_runtime_kwargs(self, runtime_configuration: Optional[dict] = None) -> dict:
        runtime_keys: tuple[str, ...]
        default_kwarg_values: Mapping[str, str | bool | float | object | None]
        try:
            impl = self._get_expectation_impl()
        except ExpectationNotFoundError:
            expectation_kwargs_dict = self._get_default_custom_kwargs()
            default_kwarg_values = expectation_kwargs_dict.get("default_kwarg_values", {})
            runtime_keys = self.runtime_kwargs
        else:
            runtime_keys = impl.runtime_keys
            default_kwarg_values = self._get_expectation_class_defaults()

        success_kwargs = self.get_success_kwargs()
        lookup_kwargs = deepcopy(self.kwargs)
        if runtime_configuration:
            lookup_kwargs.update(runtime_configuration)

        runtime_kwargs = {
            key: lookup_kwargs.get(key, default_kwarg_values.get(key)) for key in runtime_keys
        }
        runtime_kwargs["result_format"] = parse_result_format(runtime_kwargs["result_format"])
        runtime_kwargs.update(success_kwargs)

        return runtime_kwargs

    def applies_to_same_domain(
        self, other_expectation_configuration: ExpectationConfiguration
    ) -> bool:
        if self.type != other_expectation_configuration.type:
            return False
        return self.get_domain_kwargs() == other_expectation_configuration.get_domain_kwargs()

    # noinspection PyPep8Naming
    def isEquivalentTo(
        self,
        other: Union[dict, ExpectationConfiguration],
        match_type: str = "success",
    ) -> bool:
        """ExpectationConfiguration equivalence does not include meta, and relies on *equivalence* of kwargs."""  # noqa: E501 # FIXME CoP
        if not isinstance(other, self.__class__):
            if isinstance(other, dict):
                try:
                    # noinspection PyNoneFunctionAssignment
                    other = expectationConfigurationSchema.load(other)
                except ValidationError:
                    logger.debug(
                        "Unable to evaluate equivalence of ExpectationConfiguration object with dict because "  # noqa: E501 # FIXME CoP
                        "dict other could not be instantiated as an ExpectationConfiguration"
                    )
                    return NotImplemented
            else:
                # Delegate comparison to the other instance
                return NotImplemented

        if match_type == "domain":
            return all(
                (
                    self.type == other.type,  # type: ignore[union-attr] # could be dict
                    self.get_domain_kwargs() == other.get_domain_kwargs(),  # type: ignore[union-attr] # could be dict
                )
            )

        if match_type == "success":
            return all(
                (
                    self.type == other.type,  # type: ignore[union-attr] # could be dict
                    self.get_success_kwargs() == other.get_success_kwargs(),  # type: ignore[union-attr] # could be dict
                )
            )

        if match_type == "runtime":
            return all(
                (
                    self.type == other.type,  # type: ignore[union-attr] # could be dict
                    self.kwargs == other.kwargs,  # type: ignore[union-attr] # could be dict
                )
            )

        return False

    @override
    def __eq__(self, other):
        """ExpectationConfiguration equality does include meta, but ignores instance identity."""
        if not isinstance(other, self.__class__):
            # Delegate comparison to the other instance's __eq__.
            return NotImplemented
        this_kwargs: dict = convert_to_json_serializable(self.kwargs)
        other_kwargs: dict = convert_to_json_serializable(other.kwargs)
        this_meta: dict = convert_to_json_serializable(self.meta)
        other_meta: dict = convert_to_json_serializable(other.meta)
        return all(
            (
                self.type == other.type,
                this_kwargs == other_kwargs,
                this_meta == other_meta,
            )
        )

    @override
    def __hash__(self) -> int:
        this_kwargs: dict = convert_to_json_serializable(self.kwargs)
        this_meta: dict = convert_to_json_serializable(self.meta)

        def make_hashable(obj):
            """Convert unhashable types to hashable ones recursively."""
            if isinstance(obj, (str, int, float, bool, type(None))):
                return obj
            elif isinstance(obj, list):
                return tuple(make_hashable(item) for item in obj)
            elif isinstance(obj, dict):
                return tuple(sorted((k, make_hashable(v)) for k, v in obj.items()))
            else:
                return str(obj)

        return hash(
            (
                self.type,
                make_hashable(this_kwargs),
                make_hashable(this_meta),
            )
        )

    def __ne__(self, other):  # type: ignore[explicit-override] # FIXME
        # By using the == operator, the returned NotImplemented is handled correctly.
        return not self == other

    def __repr__(self):  # type: ignore[explicit-override] # FIXME
        return json.dumps(self.to_json_dict())

    @override
    def __str__(self):
        return json.dumps(self.to_json_dict(), indent=2)

    @override
    def to_json_dict(self) -> Dict[str, JSONValues]:
        """Returns a JSON-serializable dict representation of this ExpectationConfiguration.

        Returns:
            A JSON-serializable dict representation of this ExpectationConfiguration.
        """
        myself = expectationConfigurationSchema.dump(self)
        # NOTE - JPC - 20191031: migrate to expectation-specific schemas that subclass result with properly-typed  # noqa: E501 # FIXME CoP
        # schemas to get serialization all-the-way down via dump
        myself["kwargs"] = convert_to_json_serializable(myself["kwargs"])

        # Post dump hook removes this value if null so we need to ensure applicability before conversion  # noqa: E501 # FIXME CoP
        if "expectation_context" in myself:
            myself["expectation_context"] = convert_to_json_serializable(
                myself["expectation_context"]
            )
        if "rendered_content" in myself:
            myself["rendered_content"] = convert_to_json_serializable(myself["rendered_content"])
        return myself

    def _get_expectation_impl(self) -> Type[Expectation]:
        return get_expectation_impl(self.type)

    def to_domain_obj(self) -> Expectation:
        expectation_impl = self._get_expectation_impl()
        kwargs: dict[Any, Any] = {
            "id": self.id,
            "meta": self.meta,
            "notes": self.notes,
            "rendered_content": self.rendered_content,
        }
        # it's possible description could be subclassed as a class variable,
        # because we have documented it that way in the past.
        # if that is the case, passing a self.description of any type would raise an error
        # we can't check for the presence of expectation_impl.description
        # because _get_expectation_impl() only returns registered expectations
        if self.description:
            kwargs.update({"description": self.description})
        kwargs.update(self.kwargs)
        return expectation_impl(**kwargs)

    def get_domain_type(self) -> MetricDomainTypes:
        """Return "domain_type" of this expectation."""
        if self.type.startswith("expect_table_"):
            return MetricDomainTypes.TABLE

        if "column" in self.kwargs:
            return MetricDomainTypes.COLUMN

        if "column_A" in self.kwargs and "column_B" in self.kwargs:
            return MetricDomainTypes.COLUMN_PAIR

        if "column_list" in self.kwargs:
            return MetricDomainTypes.MULTICOLUMN

        raise ValueError(  # noqa: TRY003 # FIXME CoP
            'Unable to determine "domain_type" of this "ExpectationConfiguration" object from "kwargs" and heuristics.'  # noqa: E501 # FIXME CoP
        )

    def _get_expectation_class_defaults(self) -> dict[str, Any]:
        cls = self._get_expectation_impl()
        return {
            name: field.default if not field.required else None
            for name, field in cls.__fields__.items()
        }


class ExpectationConfigurationSchema(Schema):
    type = fields.Str(
        required=True,
        error_messages={"required": "type missing in expectation configuration"},
    )
    kwargs = fields.Dict(
        required=False,
        allow_none=True,
    )
    meta = fields.Dict(
        required=False,
        allow_none=True,
    )
    notes = fields.Raw(
        required=False,
        allow_none=True,
    )
    id = fields.UUID(required=False, allow_none=True)
    expectation_context = fields.Nested(
        lambda: ExpectationContextSchema,
        required=False,
        allow_none=True,
    )
    rendered_content = fields.List(
        fields.Nested(
            lambda: RenderedAtomicContentSchema,
            required=False,
            allow_none=True,
        )
    )
    description = fields.Str(required=False, allow_none=True)

    REMOVE_KEYS_IF_NONE = [
        "id",
        "expectation_context",
        "rendered_content",
        "notes",
        "description",
    ]

    @pre_dump
    def convert_result_to_serializable(self, data, **kwargs):
        data = copy.deepcopy(data)
        data["kwargs"] = convert_to_json_serializable(data.get("kwargs", {}))
        return data

    @post_dump
    def clean_null_attrs(self, data: dict, **kwargs: dict) -> dict:
        """Removes the attributes in ExpectationConfigurationSchema.REMOVE_KEYS_IF_NONE during serialization if
        their values are None."""  # noqa: E501 # FIXME CoP
        data = copy.deepcopy(data)
        for key in ExpectationConfigurationSchema.REMOVE_KEYS_IF_NONE:
            if key in data and data[key] is None:
                data.pop(key)
        return data

    def _convert_uuids_to_str(self, data):
        """
        Utilize UUID for data validation but convert to string before usage in business logic
        """
        attr = "id"
        uuid_val = data.get(attr)
        if uuid_val:
            data[attr] = str(uuid_val)
        return data

    # noinspection PyUnusedLocal
    @post_load
    def make_expectation_configuration(self, data: dict, **kwargs):
        data = self._convert_uuids_to_str(data=data)
        return ExpectationConfiguration(**data)


expectationConfigurationSchema = ExpectationConfigurationSchema()
expectationContextSchema = ExpectationContextSchema()
